import os
import json
import time
import google
from google.gax.errors import RetryError

from .model import parse_verbosity
from .util import sixdecode, retry
from . import logs


class PubsubQueue(object):
    def __init__(self, queue_name, sub_name=None, verbose=10):
        from google.cloud import pubsub

        assert 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ.keys()
        with open(os.environ['GOOGLE_APPLICATION_CREDENTIALS']) as f:
            credentials = json.loads(f.read())

        project_name = credentials['project_id']
        self.logger = logs.getLogger(self.__class__.__name__)
        if verbose is not None:
            self.logger.setLevel(parse_verbosity(verbose))

        self.pubclient = pubsub.PublisherClient()
        self.subclient = pubsub.SubscriberClient()

        self.project = project_name
        self.topic_name = self.pubclient.topic_path(project_name, queue_name)
        self.logger.info("Topic name = {}".format(self.topic_name))
        try:
            self.pubtopic = self.pubclient.get_topic(self.topic_name)
        except BaseException as e:
            self.pubtopic = self.pubclient.create_topic(self.topic_name)
            self.logger.info('topic {} created'.format(self.topic_name))

        sub_name = sub_name if sub_name else queue_name + "_sub"
        self.logger.info("Topic name = {}".format(queue_name))
        self.logger.info("Subscription name = {}".format(sub_name))

        self.sub_name = self.subclient.subscription_path(
            project_name, sub_name)
        try:
            self.subclient.get_subscription(self.sub_name)
        except BaseException as e:
            self.logger.warn(e)
            self.subclient.create_subscription(self.sub_name, self.topic_name,
                                               ack_deadline_seconds=20)

        self.logger.info('subscription {} created'.format(sub_name))

    def clean(self, timeout=0):
        while True:
            msg = self.dequeue(timeout=timeout)
            if not msg:
                break

    def get_name(self):
        return self.subclient.match_topic_from_topic_name(self.topic_name)

    def has_next(self):
        raise NotImplementedError(
            'Using has_next with distributed queue ' +
            'such as pubsub will bite you in the ass! ' +
            'Use dequeue with timeout instead')

        no_retries = 5
        for i in range(no_retries):
            response = self.subclient.api.pull(
                self.sub_name,
                return_immediately=True, max_messages=1)
            messages = response.received_messages

            retval = any(messages)
            if retval:
                break
            else:
                self.logger.debug('has_next has not found any messages, ' +
                                  'retrying (attempt {})'.format(i))
                time.sleep(5)

        for m in messages:
            self.hold(m.ack_id, 0)

        return retval

    def enqueue(self, data):
        self.logger.debug('Sending message with data \n {}'.format(data))
        data = data.encode('utf-8')
        msg_id = self.pubclient.publish(self.topic_name, data)
        self.logger.debug('Message with id {} published'.format(msg_id))

    def dequeue(self, acknowledge=True, timeout=0):
        wait_step = 1
        for waited in range(0, timeout + wait_step, wait_step):
            response = self.subclient.api.pull(
                self.sub_name,
                return_immediately=True, max_messages=1)
            msgs = response.received_messages
            if any(msgs):
                break
            elif waited == timeout:
                return None
            else:
                self.logger.info(
                    ('No messages found, sleeping for {} ' +
                     ' (total sleep time {})').format(wait_step, waited))
                time.sleep(wait_step)

        retval = msgs[0]

        if acknowledge:
            self.acknowledge(retval.ack_id)
            self.logger.debug("Message {} received and acknowledged"
                              .format(retval.message.message_id))

            return sixdecode(retval.message.data)
        else:
            self.logger.debug(
                "Message {} received, ack_id {}" .format(
                    retval.message.message_id,
                    retval.ack_id))
            return (sixdecode(retval.message.data), retval.ack_id)

    def hold(self, ack_key, delay=5):
        self.logger.debug(
            ("Message acknoledgment deadline is extended by {} " +
             "min for {}").format(
                delay,
                ack_key))
        self.subclient.modify_ack_deadline(
            self.sub_name, [ack_key], int(delay * 60))

    def acknowledge(self, ack_key):
        self.logger.debug("Message with key {} acknowledged".format(ack_key))
        retry(lambda: self.subclient.acknowledge(self.sub_name, [ack_key]),
              sleep_time=10,
              logger=self.logger,
              exception_class=RetryError)

    def delete(self):
        self.logger.debug("Deleting pubsub queue with topic" + self.topic_name)
        try:
            self.pubclient.delete_topic(self.topic_name)
            self.subclient.delete_subscription(self.sub_name)
        except google.gax.errors.RetryError as e:
            self.logger.error('Queue deletion failed with exception:')
            self.logger.exception(e)



import logging

DEBUG = logging.DEBUG
INFO = logging.INFO
ERROR = logging.ERROR

logging.basicConfig(
    format='%(asctime)s %(levelname)-6s %(name)s - %(message)s',
    level=ERROR,
    datefmt='%Y-%m-%d %H:%M:%S')


def getLogger(name):
    return logging.getLogger(name)


def debug(line):
    return logging.debug(line)


def error(line):
    return logging.error(line)


def info(line):
    return logging.info(line)



import botocore
import json
import re
from .keyvalue_provider import KeyValueProvider
from .s3_artifact_store import S3ArtifactStore


class S3Provider(KeyValueProvider):

    def __init__(self, config, blocking_auth=True, verbose=10, store=None):
        self.config = config
        self.bucket = config.get('bucket', 'studioml-meta')

        self.meta_store = S3ArtifactStore(config, verbose)

        super(S3Provider, self).__init__(
            config,
            blocking_auth,
            verbose,
            store)

    def _get(self, key, shallow=False):
        response = self.meta_store.client.list_objects_v2(
            Bucket=self.bucket,
            Prefix=key,
            Delimiter='/'
        )

        if response['KeyCount'] == 0:
            return None

        if response['KeyCount'] == 1 and \
            'Contents' in response.keys() and \
                response['Contents'][0]['Key'] == key:
            response = self.meta_store.client.get_object(
                Bucket=self.bucket,
                Key=key)
            return json.loads(response['Body'].read().decode("utf-8"))
        else:
            if response['KeyCount'] > 1:
                assert shallow, \
                    'multiple-object reads ' + \
                    'are not supported for s3 provider yet {} {}'.format(
                        key, response)

            keys = []
            keys += [c['Key'] for c in response.get('Contents', [])]
            keys += [c['Prefix'][:-1]
                     for c in response.get('CommonPrefixes', [])]
            suffixes = [re.sub('^' + key, '', k) for k in keys]

            return suffixes

    def _delete(self, key):
        response = self.meta_store.client.delete_object(
            Bucket=self.bucket,
            Key=key)

        if response['ResponseMetadata']['HTTPStatusCode'] != 204:
            raise ValueError(
                ('attempt to delete write key {} ' +
                 'returned response {}')
                .format(key, response['ResponseMetadata']))

    def _set(self, key, value):
        response = self.meta_store.client.put_object(
            Bucket=self.bucket,
            Key=key,
            Body=json.dumps(value))

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise ValueError(
                ('attempt to read write key {}  with value {} ' +
                 'returned response {}')
                .format(key, value, response['ResponseMetadata']))



import numpy as np
import pickle
import os

no_samples = 100
dim_samples = 5

learning_rate = 0.01
no_steps = 10

X = np.random.random((no_samples, dim_samples))
y = np.random.random((no_samples,))

w = np.random.random((dim_samples,))

for step in range(no_steps):
    yhat = X.dot(w)
    err = (yhat - y)
    dw = err.dot(X)
    w -= learning_rate * dw
    loss = 0.5 * err.dot(err)

    print("step = {}, loss = {}, L2 norm = {}".format(step, loss, w.dot(w)))

#    with open(os.path.expanduser('~/weights/lr_w_{}_{}.pck'
#                                 .format(step, loss)), 'w') as f:
#        f.write(pickle.dumps(w))

    from studio import fs_tracker
    with open(os.path.join(fs_tracker.get_artifact('weights'),
                           'lr_w_{}_{}.pck'.format(step, loss)),
              'w') as f:
        f.write(pickle.dumps(w))



import time
import sys
from flask import Flask, render_template, request, redirect, abort
from . import model
import argparse
import yaml
import json
import socket
import subprocess
import traceback
import six
import os
import requests
from requests.exceptions import ChunkedEncodingError

from .experiment import experiment_from_dict
from .auth import get_and_verify_user, get_auth
from . import logs

app = Flask(__name__)


DB_PROVIDER_EXPIRATION = 1800

_db_provider_timestamp = None
_db_provider = None
_config = model.get_config()

_tensorboard_dirs = {}
_save_auth_cookie = False

logger = None


@app.route('/')
def dashboard():
    return _render('dashboard.html')


@app.route('/projects')
def projects():
    return _render('projects.html')


@app.route('/users')
def users():
    return _render('users.html')


@app.route('/all')
def all_experiments():
    return _render('all_experiments.html')


@app.route('/project/<key>')
def project_details(key):
    return _render('project_details.html', project=key)


@app.route('/user/<key>')
def user_experiments(key):
    return _render("user_details.html", user=key)


@app.route('/experiment/<key>')
def experiment(key):
    return _render("experiment_details.html", experiment=key)


@app.route('/tensorboard_exp/<key>')
def tensorboard_exp(key):
    if get_allow_tensorboard():
        experiment = _db_provider.get_experiment(key, getinfo=False)
        tb_path = _db_provider.get_artifact(experiment.artifacts['tb'])

        return tensorboard(tb_path)
    else:
        return render_template(
            'error.html',
            errormsg="Tensorboard is not allowed in hosted mode yet")


@app.route('/tensorboard_proj/<key>')
def tensorboard_proj(key):
    if get_allow_tensorboard():
        experiments = [get_db().get_experiment(e) for e in
                       get_db().get_project_experiments(key)]

        logdir = ','.join(
            [e.key + ":" + get_db().get_artifact(e.artifacts['tb'])
             for e in experiments])

        return tensorboard(logdir)
    else:
        return render_template(
            'error.html',
            errormsg="TensorBoard is not allowed in hosted mode yet")


def tensorboard(logdir):
    port = _tensorboard_dirs.get(logdir)
    if not port:

        sock = socket.socket(socket.AF_INET)
        sock.bind(('', 0))
        port = sock.getsockname()[1]
        sock.close()

        subprocess.Popen([
            'tensorboard',
            '--logdir=' + logdir,
            '--port=' + str(port)])
        time.sleep(5)  # wait for tensorboard to spin up
        _tensorboard_dirs[logdir] = port

    redirect_url = 'http://{}:{}'.format(
        six.moves.urllib.parse.urlparse(request.url).hostname,
        port)

    logger.debug('Redirecting to ' + redirect_url)
    return redirect(redirect_url)


@app.route('/api/get_experiment', methods=['POST'])
def get_experiment():
    tic = time.time()
    key = request.json['key']
    get_urls = request.json.get('get_artifact_urls', True)
    getlogger().info('Getting experiment {} '.format(key))
    try:
        experiment = get_db().get_experiment(key)
        if get_urls:
            artifacts = get_db().get_artifacts(experiment)
            for art, url in six.iteritems(artifacts):
                experiment.artifacts[art]['url'] = url

        experiment_data = experiment.__dict__
        status = 'ok'
    except BaseException:
        experiment_data = {}
        status = traceback.format_exc()

    retval = json.dumps({
        'status': status,
        'experiment': experiment_data
    })

    toc = time.time()
    getlogger().info('Processed get_experiment request in {} s'
                     .format(toc - tic))

    return retval


@app.route('/api/get_user_experiments', methods=['POST'])
def get_user_experiments():
    tic = time.time()
    myuser_id = get_and_verify_user(request, get_auth_config())
    if request.json and 'user' in request.json.keys():
        user = request.json['user']
    else:
        user = myuser_id

    # TODO check is myuser_id is authorized to do that

    getlogger().info('Getting experiments of user {}'
                     .format(user))

    experiments = [
        e for e in get_db().get_user_experiments(
            user, blocking=True)]
    status = "ok"
    retval = json.dumps({
        "status": status,
        "experiments": experiments
    })
    toc = time.time()
    getlogger().info('Processed get_user_experiments request in {} s'
                     .format(toc - tic))
    return retval


@app.route('/api/get_all_experiments', methods=['POST'])
def get_all_experiments():
    tic = time.time()
    get_and_verify_user(request, get_auth_config())

    # TODO check is myuser_id is authorized to do that

    getlogger().info('Getting all experiments')
    users = get_db().get_users()

    experiments = [e for user in users
                   for e in get_db().get_user_experiments(
                       user, blocking=False)]

    status = "ok"
    retval = json.dumps({
        "status": status,
        "experiments": experiments
    })
    toc = time.time()
    getlogger().info('Processed get_user_experiments request in {} s'
                     .format(toc - tic))
    return retval


@app.route('/api/get_projects', methods=['POST'])
def get_projects():
    tic = time.time()
    get_and_verify_user(request, get_auth_config())

    # TODO check / filter access

    projects = get_db().get_projects()
    status = "ok"

    retval = json.dumps({
        "status": status,
        "projects": projects
    })

    toc = time.time()
    getlogger().info('Processed get_projects request in {} s'
                     .format(toc - tic))
    return retval


@app.route('/api/get_users', methods=['POST'])
def get_users():
    tic = time.time()
    get_and_verify_user(request, get_auth_config())

    # TODO check / filter access

    users = get_db().get_users()
    status = "ok"

    retval = json.dumps({
        "status": status,
        "users": users
    })
    toc = time.time()
    getlogger().info('Processed get_user_experiments request in {} s'
                     .format(toc - tic))
    return retval


@app.route('/api/get_project_experiments', methods=['POST'])
def get_project_experiments():
    tic = time.time()
    get_and_verify_user(request, get_auth_config())

    project = request.json.get('project')
    if not project:
        status = "Project is none!"
        experiments = []
    else:
        # TODO check is myuser_id is authorized to do that

        getlogger().info('Getting experiments in project {}'
                         .format(project))

        experiments = get_db().get_project_experiments(project)

    status = "ok"
    retval = json.dumps({
        "status": status,
        "experiments": experiments
    })
    toc = time.time()
    getlogger().info('Processed get_project_experiments request in {} s'
                     .format(toc - tic))
    return retval


@app.route('/api/delete_experiment', methods=['POST'])
def delete_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())
    try:
        key = request.json['key']
        if get_db().can_write_experiment(key, userid):
            getlogger().info('Deleting experiment {} '.format(key))
            get_db().delete_experiment(key)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException:
        status = traceback.format_exc()

    toc = time.time()
    getlogger().info('Processed delete_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status})


@app.route('/api/stop_experiment', methods=['POST'])
def stop_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())
    try:
        key = request.json['key']
        if get_db().can_write_experiment(key, userid):
            getlogger().info('Stopping experiment {} '.format(key))
            get_db().stop_experiment(key)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException as e:
        status = e.message

    toc = time.time()
    getlogger().info('Processed stop_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status})


@app.route('/api/start_experiment', methods=['POST'])
def start_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())
    try:
        key = request.json['key']
        if get_db().can_write_experiment(key, userid):
            getlogger().info('Starting experiment {} '.format(key))
            experiment = get_db().get_experiment(key)
            get_db().start_experiment(experiment)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException as e:
        status = e.message

    toc = time.time()
    getlogger().info('Processed start_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status})


@app.route('/api/finish_experiment', methods=['POST'])
def finish_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())
    try:
        key = request.json['key']
        if get_db().can_write_experiment(key, userid):
            getlogger().info('Finishing experiment {} '.format(key))
            get_db().finish_experiment(key)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException as e:
        status = e.message

    toc = time.time()
    getlogger().info('Processed start_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status})


@app.route('/api/add_experiment', methods=['POST'])
def add_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())

    artifacts = {}
    try:
        experiment = experiment_from_dict(request.json['experiment'])
        compression = request.json.get('compression')
        compression = compression if compression else 'bzip2'

        if get_db().can_write_experiment(experiment.key, userid):
            for tag, art in six.iteritems(experiment.artifacts):
                art.pop('local', None)

            get_db().add_experiment(experiment, userid,
                                    compression=compression)
            added_experiment = get_db().get_experiment(experiment.key)

            artifacts = _process_artifacts(added_experiment)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException:
        status = traceback.format_exc()
    toc = time.time()
    getlogger().info('Processed add_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status, 'artifacts': artifacts})


@app.route('/api/checkpoint_experiment', methods=['POST'])
def checkpoint_experiment():
    tic = time.time()
    userid = get_and_verify_user(request, get_auth_config())

    artifacts = {}
    try:
        key = request.json['key']
        if get_db().can_write_experiment(key, userid):
            experiment = get_db().get_experiment(key)
            get_db().checkpoint_experiment(experiment)

            artifacts = _process_artifacts(experiment)
            status = 'ok'
        else:
            raise ValueError('Unauthorized')

    except BaseException:
        status = traceback.format_exc()

    toc = time.time()
    getlogger().info('Processed checkpoint_experiment request in {} s'
                     .format(toc - tic))

    return json.dumps({'status': status, 'artifacts': artifacts})


@app.route('/api/exchange_github_code')
def exchange_github_code():
    tic = time.time()
    code = request.args.get('code')

    getlogger().debug('Code = ' + code)

    try:
        response = requests.post(
            'https://github.com/login/oauth/access_token',
            json={
                'client_id': os.environ.get('STUDIO_GITHUB_ID'),
                'client_secret': os.environ.get('STUDIO_GITHUB_SECRET'),
                'code': code,
            })
    except ChunkedEncodingError as e:
        getlogger().info(e.__dict__)
        raise e

    if response.status_code != 200:
        abort(response.status_code)
    else:
        getlogger().info(response.content)

        toc = time.time()
        getlogger().info('Processed exchange_github_code in {} s'
                         .format(toc - tic))

        return response.content


def _process_artifacts(experiment):
    artifacts = {}
    for tag, art in six.iteritems(experiment.artifacts):
        if 'key' in art.keys():
            put_url, timestamp = get_db().store.get_artifact_url(
                art, method='PUT', get_timestamp=True)

            art['url'] = put_url
            art['timestamp'] = timestamp
            artifacts[tag] = art

    return artifacts


def get_db():
    global _config
    global _db_provider
    global _db_provider_timestamp

    if not _db_provider or \
       not _db_provider_timestamp or \
            time.time() - _db_provider_timestamp > DB_PROVIDER_EXPIRATION:
        _db_provider = model.get_db_provider(_config, blocking_auth=False)
        _db_provider_timestamp = time.time()

    return _db_provider


def get_allow_tensorboard():
    global _save_auth_cookie
    return _save_auth_cookie


def getlogger():
    global logger
    if logger is None:
        logger = logs.getLogger('studio_server')
        logger.setLevel(10)

    return logger


def get_config():
    global _config
    if _config is None:
        _config = model.get_config()
    return _config


def get_auth_config():
    return get_config()['server']['authentication']


def _render(page, **kwargs):
    tic = time.time()
    token = None
    if _save_auth_cookie:
        auth = get_auth(get_auth_config())
        if auth:
            token = auth.get_token()

    retval = render_template(
        page,
        api_key=get_db().app.api_key,
        project_id=_config['database'].get('project_id'),
        send_refresh_token="true",
        allow_tensorboard=get_allow_tensorboard(),
        github_client_id=os.environ.get('STUDIO_GITHUB_ID'),
        auth_token=token,
        **kwargs
    )
    toc = time.time()
    getlogger().info('page {} rendered in {} s'.
                     format(page, toc - tic))
    return retval


def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description='Studio WebUI server. \
                     Usage: studio \
                     <arguments>')

    parser.add_argument('--config', help='configuration file', default=None)
#    parser.add_argument('--guest',
#                        help='Guest mode (does not require db credentials)',
#                        action='store_true')

    parser.add_argument('--port',
                        help='port to run Flask server on',
                        type=int,
                        default=5000)

    parser.add_argument('--host',
                        help='host name.',
                        default='localhost')

    parser.add_argument(
        '--verbose', '-v',
        help='Verbosity level. Allowed vaules: ' +
             'debug, info, warn, error, crit ' +
             'or numerical value of logger levels.',
        default=None)

    args = parser.parse_args(args)
    config = model.get_config()
    if args.config:
        with open(args.config) as f:
            config = yaml.load(f)

    if args.verbose:
        config['verbose'] = args.verbose

#    if args.guest:
#        config['database']['guest'] = True
    global _config
    global _db_provider
    _config = config
    _db_provider = model.get_db_provider(_config)

    getlogger().setLevel(model.parse_verbosity(config.get('verbose')))

    global _save_auth_cookie
    _save_auth_cookie = True

    print('Starting Studio UI on port {0}'.format(args.port))
    app.run(host=args.host, port=args.port)


if __name__ == "__main__":
    main()



import argparse
import sys
import os
import time
import json
import glob
import traceback
import importlib
import pickle
import re
import threading

from flask import Flask, request
from . import fs_tracker, logs
from .model_util import ModelPipe

app = Flask(__name__)
model = None
logger = None

killtimer = None
killtimer_duration = None


def get_logger():
    global logger
    if not logger:
        logger = logs.getLogger('studio-serve')
        logger.setLevel(logs.DEBUG)
    return logger


def restart_killtimer(duration=None):
    global killtimer
    global killtimer_duration

    def killtriger():
        get_logger().info('Shutting down due to inactivity')
        os._exit(0)

    if killtimer:
        killtimer.cancel()

    if duration:
        killtimer_duration = duration
    else:
        duration = killtimer_duration

    get_logger().info('Initializing kill timer for {} s'
                      .format(duration))
    killtimer = threading.Timer(duration, killtriger)
    killtimer.start()


@app.route('/', methods=['POST'])
def inference():
    try:
        tic = time.time()
        input_dict = request.json
        output_dict = model(input_dict)
        return json.dumps(output_dict)
    except BaseException as e:
        return json.dumps({
            'error': traceback.format_exc(e)
        })
    finally:
        get_logger().info('inference completed in {} s'
                          .format(time.time() - tic))

        restart_killtimer()


def main():
    argparser = argparse.ArgumentParser(
        description='Serve studio model'
    )

    argparser.add_argument(
        '--wrapper', '-w',
        help='python script with function create_model ' +
             'that takes modeldir '
        '(that is, directory where experiment saves ' +
        'the checkpoints etc)' +
        'and returns dict -> dict function (model).' +
        'By default, studio-serve will try to determine ' +
        'this function automatically.',
        default=None
    )

    argparser.add_argument('--port',
                           help='port to run Flask server on',
                           type=int,
                           default=5000)

    argparser.add_argument('--host',
                           help='host name.',
                           default='0.0.0.0')

    argparser.add_argument(
        '--killafter',
        help='Shut down after this many seconds of inactivity',
        default=3600)

    options = argparser.parse_args(sys.argv[1:])

    global model

    modeldir = fs_tracker.get_artifact('modeldata')
    if options.wrapper:
        module_name = re.sub('.py\Z', '', options.wrapper)
        wrapper_module = importlib.import_module(module_name)
        model = wrapper_module.create_model(modeldir)
    else:
        model = auto_generate_model(modeldir)

    restart_killtimer(int(options.killafter))
    app.run(host=options.host, port=options.port)


def auto_generate_model(modeldir):
    if modeldir is not None:
        hdf5_files = [
            (p, os.path.getmtime(p))
            for p in
            glob.glob(modeldir + '/*.hdf*') +
            glob.glob(modeldir + '/*.h5')]
        if any(hdf5_files):
            # experiment type - keras
            get_logger().info("Loading keras model " +
                              "(using pickle serialization)")
            import keras
            last_checkpoint = max(hdf5_files, key=lambda t: t[1])[0]
            keras_model = keras.models.load_model(last_checkpoint)
            print(keras_model)
            keras_model.summary()
            return wrap_keras_model(keras_model)

    return lambda x: x


def wrap_keras_model(keras_model):
    pipe = ModelPipe()
    pipe.add(lambda d: pickle.loads(d))
    pipe.add(keras_model)
    pipe.add(lambda d: pickle.dumps(d))

    return pipe


if __name__ == '__main__':
    main()



import threading

from datetime import timedelta

from cachetools import LRUCache
from .rabbit_queue import RMQueue

queue_cache = {}


def get_cached_queue(
        name,
        route,
        cloud=None,
        config=None,
        logger=None,
        close_after=timedelta(),
        verbose=10):
    q = RMQueue(
        queue=name,
        route=route,
        config=config,
        logger=logger,
        verbose=verbose)
    if close_after is not None and close_after.total_seconds() > 0:
        thr = threading.Timer(
            interval=close_after.total_seconds(),
            function=purge_rmq,
            kwargs={
                "q": q,
                "logger": logger})
        thr.setDaemon(True)
        thr.start()

    return q


def purge_rmq(q, logger, **kwargs):
    if q is None:
        return

    try:
        q.stop()
    except BaseException as e:
        logger.warning(e)
        return
    return



import unittest
import yaml
import uuid
import os
import time
import tempfile
import shutil
import requests
import subprocess
import boto3

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from studio import model
from studio.auth import remove_all_keys

from studio.gcloud_artifact_store import GCloudArtifactStore
from studio.s3_artifact_store import S3ArtifactStore

from studio.util import has_aws_credentials
from env_detect import on_gcp, on_aws


class ArtifactStoreTest(object):
    _multiprocess_shared_ = True

    def get_store(self, config_name='test_config.yaml'):
        config_file = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)),
            config_name)
        with open(config_file) as f:
            config = yaml.load(f)

        return model.get_db_provider(config).store

    def test_get_put_artifact(self):
        fb = self.get_store()
        tmp_dir = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        random_str = str(uuid.uuid4())

        os.makedirs(os.path.join(tmp_dir, 'test_dir'))

        tmp_filename = os.path.join(
            tmp_dir, 'test_dir', str(uuid.uuid4()) + '.txt')
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        artifact = {'key': 'tests/' + str(uuid.uuid4()) + '.tar'}
        fb.put_artifact(artifact, tmp_dir, cache=False)
        shutil.rmtree(tmp_dir)
        fb.get_artifact(artifact, tmp_dir)

        with open(tmp_filename, 'r') as f:
            line = f.read()

        shutil.rmtree(tmp_dir)
        self.assertTrue(line == random_str)
        fb.delete_artifact(artifact)

    @unittest.skip('sometimes fails in travis at time assertion' +
                   ', peterz figure out')
    def test_get_put_cache(self):
        fb = self.get_store()
        tmp_dir = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        strlen = 10000000
        random_str = str(os.urandom(strlen))

        os.makedirs(os.path.join(tmp_dir, 'test_dir'))
        tmp_filename = os.path.join(
            tmp_dir, 'test_dir', str(uuid.uuid4()) + '.txt')

        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        artifact = {'key': 'tests/' + str(uuid.uuid4()) + '.tgz'}

        fb.put_artifact(artifact, tmp_dir, cache=False)
        shutil.rmtree(tmp_dir)

        tic1 = time.clock()
        fb.get_artifact(artifact, tmp_dir)
        os.utime(tmp_dir, None)
        tic2 = time.clock()
        fb.get_artifact(artifact, tmp_dir)
        tic3 = time.clock()

        with open(tmp_filename, 'r') as f:
            line = f.read()

        shutil.rmtree(tmp_dir)
        self.assertTrue(line == random_str)

        self.assertTrue(tic3 - tic2 < (tic2 - tic1))
        fb.delete_artifact(artifact)

    def test_get_artifact_url(self):
        remove_all_keys()
        fb = self.get_store('test_config.yaml')
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        artifact_key = 'tests/test_' + str(uuid.uuid4())
        artifact = {'key': artifact_key}
        fb.put_artifact(artifact, tmp_filename, cache=False)
        url = fb.get_artifact_url(artifact)
        os.remove(tmp_filename)
        response = requests.get(url)
        self.assertEquals(response.status_code, 200)
        tar_filename = os.path.join(tempfile.gettempdir(),
                                    str(uuid.uuid4()) + '.tgz')
        with open(tar_filename, 'wb') as f:
            f.write(response.content)

        ptar = subprocess.Popen(['tar', '-xf', tar_filename],
                                cwd=tempfile.gettempdir())

        tarout, _ = ptar.communicate()

        with open(tmp_filename, 'r') as f:
            self.assertEquals(f.read(), random_str)

        os.remove(tmp_filename)
        os.remove(tar_filename)
        fb.delete_artifact(artifact)

    def test_delete_file(self):
        fb = self.get_store()
        tmp_dir = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        random_str = str(uuid.uuid4())

        os.makedirs(os.path.join(tmp_dir, 'test_dir'))
        tmp_filename = os.path.join(
            tmp_dir, 'test_dir', str(uuid.uuid4()) + '.txt')
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        artifact_key = 'tests/test_' + str(uuid.uuid4())
        artifact = {'key': artifact_key}
        fb.put_artifact(artifact, tmp_filename, cache=False)
        shutil.rmtree(tmp_dir)
        fb.delete_artifact(artifact)
        fb.get_artifact(artifact, tmp_dir)

        exception_raised = False
        try:
            with open(tmp_filename, 'r') as f:
                f.read()
        except IOError:
            exception_raised = True

        self.assertTrue(exception_raised)

    def test_get_qualified_location(self):
        fb = self.get_store()
        key = str(uuid.uuid4())
        qualified_location = fb.get_qualified_location(key)
        expected_qualified_location = self.get_qualified_location_prefix() + \
            key

        self.assertEquals(qualified_location, expected_qualified_location)


class FirebaseArtifactStoreTest(ArtifactStoreTest, unittest.TestCase):
    # Tests of private methods

    def get_qualified_location_prefix(self):
        return "gs://studio-ed756.appspot.com/"

    def test_get_file_url(self):
        remove_all_keys()
        fb = self.get_store('test_config.yaml')

        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'wt') as f:
            f.write(random_str)

        key = 'tests/' + str(uuid.uuid4()) + '.txt'
        fb._upload_file(key, tmp_filename)

        url = fb._get_file_url(key)
        response = requests.get(url)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content.decode('utf-8'),
                          random_str)
        fb._delete_file(key)
        os.remove(tmp_filename)

    def test_get_file_url_auth(self):
        remove_all_keys()
        fb = self.get_store('test_config_auth.yaml')
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')
        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        key = 'authtest/' + str(uuid.uuid4()) + '.txt'
        fb._upload_file(key, tmp_filename)

        url = fb._get_file_url(key)
        response = requests.get(url)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content.decode('utf-8'),
                          random_str)
        fb._delete_file(key)
        os.remove(tmp_filename)

    def test_upload_download_file_noauth(self):
        remove_all_keys()
        fb = self.get_store('test_config.yaml')
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        key = 'authtest/' + str(uuid.uuid4()) + '.txt'
        fb._upload_file(key, tmp_filename)
        os.remove(tmp_filename)
        fb._download_file(key, tmp_filename)

        self.assertTrue(not os.path.exists(tmp_filename))

    def test_upload_download_file_bad(self):
        # smoke test to make sure attempt to access a wrong file
        # in the database is wrapped and does not crash the system
        fb = self.get_store('test_bad_config.yaml')
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            'test_upload_download.txt')
        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        fb._upload_file('tests/test_upload_download.txt', tmp_filename)
        fb._download_file('tests/test_upload_download.txt', tmp_filename)
        os.remove(tmp_filename)

    def test_upload_download_file_auth(self):
        remove_all_keys()
        fb = self.get_store('test_config_auth.yaml')
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        key = 'authtest/' + str(uuid.uuid4())
        fb._upload_file(key, tmp_filename)
        os.remove(tmp_filename)

        # test an authorized attempt to delete file
        remove_all_keys()
        fb = self.get_store('test_config.yaml')
        fb._delete_file(key)
        remove_all_keys()
        fb = self.get_store('test_config_auth.yaml')
        # to make sure file is intact and the same as we uploaded

        fb._download_file(key, tmp_filename)

        with open(tmp_filename, 'r') as f:
            line = f.read()
        os.remove(tmp_filename)
        self.assertTrue(line == random_str)

        fb._delete_file(key)
        fb._download_file(key, tmp_filename)
        self.assertTrue(not os.path.exists(tmp_filename))

    def test_upload_download_file(self):
        fb = self.get_store()
        tmp_filename = os.path.join(
            tempfile.gettempdir(),
            str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmp_filename, 'w') as f:
            f.write(random_str)

        key = 'tests/' + str(uuid.uuid4()) + '.txt'
        fb._upload_file(key, tmp_filename)
        os.remove(tmp_filename)
        fb._download_file(key, tmp_filename)
        with open(tmp_filename, 'r') as f:
            line = f.read()
        os.remove(tmp_filename)
        self.assertTrue(line == random_str)

        fb._delete_file(key)
        fb._download_file(key, tmp_filename)
        self.assertTrue(not os.path.exists(tmp_filename))


@unittest.skipIf(
    not on_gcp(),
    'User indicated not on gcp')
class UserIndicatedOnGCPTest(unittest.TestCase):
    def test_on_enviornment(self):
        self.assertTrue('GOOGLE_APPLICATION_CREDENTIALS' in os.environ.keys())


@unittest.skipIf(
    (not on_gcp()) or
    'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
    'Skipping due to userinput or GCP Not detected')
class GCloudArtifactStoreTest(ArtifactStoreTest, unittest.TestCase):

    def get_store(self, config_name=None):
        store = ArtifactStoreTest.get_store(
            self, 'test_config_gcloud_storage.yaml')
        self.assertTrue(isinstance(store, GCloudArtifactStore))
        return store

    def get_qualified_location_prefix(self):
        store = self.get_store()
        return "gs://" + store.get_bucket() + "/"


@unittest.skipIf(
    not on_aws(),
    'User indicated not on aws')
class UserIndicatedOnAWSTest(unittest.TestCase):
    def test_on_enviornment(self):
        self.assertTrue(has_aws_credentials())


@unittest.skipIf(
    (not on_aws()) or not has_aws_credentials(),
    'Skipping due to userinput or AWS Not detected')
class S3ArtifactStoreTest(ArtifactStoreTest, unittest.TestCase):

    def get_store(self, config_name=None):
        store = ArtifactStoreTest.get_store(
            self, 'test_config_s3_storage.yaml')
        self.assertTrue(isinstance(store, S3ArtifactStore))
        return store

    def get_qualified_location_prefix(self):
        store = self.get_store()
        endpoint = urlparse(boto3.client('s3')._endpoint.host)
        return "s3://" + endpoint.netloc + "/" + store.bucket + "/"


if __name__ == "__main__":
    unittest.main()



import hashlib
import six
from studio import util


def clientFunction(args, files):
    print('client function call with args ' +
          str(args) + ' and files ' + str(files))

    cs_files = {'output', 'clientscript', 'args'}
    filehashes = {
        k: util.filehash(
            v,
            hashobj=hashlib.md5()) for k,
        v in six.iteritems(files) if k not in cs_files}

    return (args, filehashes)



from .firebase_artifact_store import FirebaseArtifactStore
from .gcloud_artifact_store import GCloudArtifactStore
from .local_artifact_store import LocalArtifactStore
from .s3_artifact_store import S3ArtifactStore

def get_artifact_store(config, blocking_auth=True, verbose=10):
    if config['type'].lower() == 'firebase':
        return FirebaseArtifactStore(
            config, blocking_auth=blocking_auth, verbose=verbose)
    elif config['type'].lower() == 'gcloud':
        return GCloudArtifactStore(config, verbose=verbose)
    elif config['type'].lower() == 's3':
        return S3ArtifactStore(config, verbose=verbose)
    elif config['type'].lower() == 'local':
        return LocalArtifactStore(config, verbose=verbose)
    else:
        raise ValueError('Unknown storage type: ' + config['type'])



import sys
import argparse
from . import runner


def main():
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        '--wrapper', '-w',
        help='python script with function create_model ' +
             'that takes modeldir '
        '(that is, directory where experiment saves ' +
        'the checkpoints etc)' +
        'and returns dict -> dict function (model).' +
        'By default, studio-serve will try to determine ' +
        'this function automatically.',
        default=None
    )

    argparser.add_argument('--port',
                           help='port to run Flask server on',
                           type=int,
                           default=5000)

    argparser.add_argument('--host',
                           help='host name.',
                           default='0.0.0.0')

    argparser.add_argument(
        '--killafter',
        help='Shut down after this many seconds of inactivity',
        default=3600)

    options, other_args = argparser.parse_known_args(sys.argv[1:])
    serve_args = ['studio::serve_main']

    assert len(other_args) >= 1
    experiment_key = other_args[-1]
    runner_args = other_args[:-1]
    runner_args.append('--reuse={}/modeldir:modeldata'.format(experiment_key))
    runner_args.append('--force-git')
    runner_args.append('--port=' + str(options.port))

    if options.wrapper:
        serve_args.append('--wrapper=' + options.wrapper)
        serve_args.append('--port=' + str(options.port))

    serve_args.append('--host=' + options.host)
    serve_args.append('--killafter=' + str(options.killafter))

    total_args = runner_args + serve_args
    runner.main(total_args)


if __name__ == '__main__':
    main()



import unittest
from studio.gpu_util import memstr2int


class GpuUtilTest(unittest.TestCase):

    def test_memstr2int(self):
        self.assertEquals(memstr2int('123 Mb'), 123 * (2**20))
        self.assertEquals(memstr2int('456 MiB'), 456 * (2**20))
        self.assertEquals(memstr2int('23 Gb'), 23 * (2**30))
        self.assertEquals(memstr2int('23 GiB'), 23 * (2**30))
        self.assertEquals(memstr2int('23 '), 23)

        with self.assertRaises(ValueError):
            memstr2int('300 spartans')


if __name__ == "__main__":
    unittest.main()



import types
try:
    import keras
    from keras.preprocessing import image as image_prep
except ImportError:
    keras = None

from . import logs

from PIL import Image

try:
    from Queue import Full, Empty, Queue
except ImportError:
    from queue import Full, Empty, Queue

from threading import Thread
import numpy as np
import itertools
import six


class BufferedPipe:
    def __init__(self,
                 func=lambda x: x,
                 parent=None,
                 q_in=None,
                 q_out=None,
                 num_workers=0,
                 q_size=None,
                 batch_size=1,
                 filterf=lambda x: x is not None,
                 batcher=lambda x: x,
                 timeout=1):

        min_q_size = 10

        self.func = func
        self.parent = parent
        self.num_workers = num_workers
        self.filterf = filterf
        self.batch_size = batch_size
        self.batcher = batcher

        if num_workers > 0:
            self.q_size = q_size if q_size else 2 * num_workers

        self.q_out = q_out
        self.q_in = q_in
        self.q_size = max(min_q_size, 2 * num_workers)

        self.logger = logs.getLogger('BufferedPipe')
        self.logger.setLevel(10)
        self.timeout = timeout
        self.worker_frame = Thread

    def __call__(self, data_gen):
        if self.parent:
            data_gen = self.parent(data_gen)

        if self.num_workers == 0 and \
           self.batch_size == 1 and \
           self.q_in is None and \
           self.q_out is None:
            return (self._wrapped_func(x) for x in data_gen)

        q_in = self.q_in
        if q_in is None:
            q_in = Queue(self.q_size)
            Thread(target=_gen2q,
                   args=(data_gen, q_in),
                   ).start()

        q_out = self.q_out
        if q_out is None:
            q_out = Queue(self.q_size)

        if self.batch_size == 1:
            target = _q2q_single
            kwargs = {
                "func": self._wrapped_func,
                "queue_in": q_in,
                "queue_out": q_out,
                "filterf": self._wrapped_filter,
                "timeout": self.timeout}
        else:
            target = _q2q_batch
            kwargs = {
                "func": self._wrapped_func,
                "queue_in": q_in,
                "queue_out": q_out,
                "filterf": self._wrapped_filter,
                "batch_size": self.batch_size,
                "timeout": self.timeout}

        if self.num_workers == 0:
            self.worker_frame(target=target, kwargs=kwargs).start()
        else:
            for i in range(self.num_workers):
                self.worker_frame(target=target,
                                  kwargs=kwargs).start()

        if self.q_out is None:
            return _q2gen(q_out, timeout=self.timeout)

    def add(self, func,
            num_workers=None,
            batch_size=None,
            filterf=None,
            batcher=None,
            timeout=None):

        if num_workers is None and \
           batch_size is None and \
           filterf is None and \
           batcher is None and \
           timeout is None:
            g = self.func
            self.func = lambda x: func(g(x))
            return self
        else:
            assert self.q_out is None
            self.q_out = Queue(self.q_size)
            return BufferedPipe(
                func, self,
                q_in=self.q_out,
                num_workers=num_workers if num_workers else self.num_workers,
                batch_size=batch_size if batch_size else self.batch_size,
                filterf=filterf if filterf else self.filterf,
                batcher=batcher if batcher else self.batcher,
                timeout=timeout if timeout else self.timeout)

    def _wrapped_func(self, x):
        if isinstance(x, tuple):
            try:
                return (x[0], self.func(x[1]))
            except BaseException as e:
                self.logger.warn('Applying function to {} raised exception {}'
                                 .format(x[1], str(e)))
                self.logger.exception(e)
                return (x[0], None)

        elif isinstance(x, list) and isinstance(x[0], tuple):
            batch_index = [el[0] for el in x]
            batch_input = self.batcher([el[1] for el in x])
            try:
                batch_output = self.func(batch_input)
            except BaseException as e:
                self.logger.warn('Applying function to {} raised exception {}'
                                 .format(batch_input, str(e)))
                self.logger.exception(e)
                batch_output = [None] * len(batch_index)

            try:
                return zip(batch_index, batch_output)
            except BaseException as e:
                self.logger.warn('Applying function to {} raised exception {}'
                                 .format(x, str(e)))
                self.logger.exception(e)
                return None

        else:
            try:
                return self.func(x)
            except BaseException as e:
                self.logger.warn('Applying function to {} raised exception {}'
                                 .format(x, str(e)))
                self.logger.exception(e)
                return None

    def _wrapped_filter(self, x):
        if isinstance(x, tuple):
            return self.filterf(x[1])
        else:
            return self.filterf(x)


class ModelPipe:
    def __init__(self):
        self._pipe = BufferedPipe()

    def add(self, func,
            num_workers=None,
            batch_size=None,
            filterf=None,
            batcher=None,
            timeout=None):
        if keras and (isinstance(func, keras.models.Sequential) or
                      isinstance(func, keras.models.Model)):
            model = func
            _prime_keras_model(func)
            func = model.predict

        self._pipe = self._pipe.add(
            func,
            num_workers,
            batch_size=batch_size,
            filterf=filterf,
            batcher=batcher,
            timeout=timeout)
        return self

    def apply_unordered(self, data):

        if not isinstance(data, dict):
            count_gen = itertools.count(start=0, step=1)
            indexed_gen = six.moves.zip(count_gen, (x for x in data))
        else:
            indexed_gen = ((k, v) for k, v in six.iteritems(data))

        output_gen = self._pipe(indexed_gen)

        if isinstance(data, list):
            return [x for x in output_gen]
        elif isinstance(data, types.GeneratorType):
            return output_gen
        elif isinstance(data, dict):
            return {x[0]: x[1] for x in output_gen}
        elif isinstance(data, set):
            return {x[1] for x in output_gen}

    def apply_ordered(self, data):
        unordered = self.apply_unordered(data)
        if isinstance(data, dict) or \
           isinstance(data, set):
            return unordered
        elif isinstance(data, list):
            return [x[1] for x in sorted(unordered, key=lambda x:x[0])]
        elif isinstance(data, types.GeneratorType):
            return (x[1] for x in sorted(unordered, key=lambda x: x[0]))

    def __call__(self, data):
        return self.apply_unordered(data)


def resize_to_model_input(model, input_index=0):
    assert keras is not None
    assert isinstance(
        model, keras.models.Model) or isinstance(
        model, keras.models.Sequential)

    input_shape = tuple(
        [x.value for x in model.inputs[input_index].shape if x.value])
    assert len(input_shape) == 3

    if len(input_shape) == 3:
        assert input_shape[0] == 1 or input_shape[2] == 1 or \
            input_shape[0] == 3 or input_shape[2] == 3

        if input_shape[0] == 1 or input_shape[0] == 3:
            data_format = 'channels_first'
        else:
            data_format = 'channels_last'

    def _run_resize(input_img):
        if input_img is None:
            return None
        if len(input_shape) == 3:
            img = input_img.resize(
                (input_shape[1], input_shape[0]), Image.ANTIALIAS)

        arr = image_prep.img_to_array(img, data_format)
        arr /= 255.0

        if input_shape[0] == 1 and arr.shape[0] > 1:
            arr = np.mean(arr, axis=0)

        if input_shape[2] == 1 and arr.shape[2] > 1:
            arr = np.mean(arr, axis=2)

        return arr.reshape((1,) + input_shape)

    return _run_resize


def _q2q_batch(
        func,
        queue_in,
        queue_out,
        filterf=lambda x: x is not None,
        batch_size=32,
        timeout=1):
    while True:
        added = 0
        batch = []
        while True:
            try:
                if added > 0:
                    next_el = queue_in.get_nowait()
                else:
                    next_el = queue_in.get(True, timeout)

                if filterf(next_el):
                    batch.append(next_el)
                    added += 1

                if added == batch_size:
                    break
            except Empty:
                break

        if not any(batch):
            return

        retval = func(batch)

        for el in retval:
            put_success = False
            while not put_success:
                try:
                    queue_out.put_nowait(el)
                    put_success = True
                except Full:
                    pass


def _q2q_single(func, queue_in, queue_out, filterf=lambda x: True, timeout=1):
    _q2q_batch(
        lambda b: [
            func(x) for x in b],
        queue_in,
        queue_out,
        filterf,
        1,
        timeout)


def _gen2q(data, queue):
    while True:
        try:
            next_el = next(data)
        except StopIteration:
            return

        enqueued_successfully = False
        while not enqueued_successfully:
            try:
                queue.put_nowait(next_el)
                enqueued_successfully = True
            except Full:
                pass


def _q2gen(queue, timeout=1):
    while True:
        try:
            yield queue.get(True, timeout=timeout)
        except Empty:
            raise StopIteration


def _prime_keras_model(model):
    input_shapes = [[s.value if s.value else 1 for s in l.shape]
                    for l in model.inputs]
    dummy_inputs = [np.random.random(shape) for shape in input_shapes]

    model.predict(dummy_inputs)



import argparse
import sys
import time
from terminaltables import AsciiTable

from . import model
from . import logs

_my_logger = None


def print_help():
    print('Usage: studio runs [command] arguments')
    print('\ncommand can be one of the following:')
    print('')
    print('\tlist [username] - display experiments')
    print('\tstop [experiment] - stop running experiment')
    print('\tkill [experiment] - stop and delete experiment')
    print('')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', help='configuration file', default=None)
    parser.add_argument(
        '--short', '-s', help='Brief output - names of experiments only',
        action='store_true')

    cli_args, script_args = parser.parse_known_args(sys.argv)

    get_logger().setLevel(10)

    if len(script_args) < 2:
        get_logger().critical('No command provided!')
        parser.print_help()
        print_help()
        return

    cmd = script_args[1]

    if cmd == 'list':
        _list(script_args[2:], cli_args)
    elif cmd == 'stop':
        _stop(script_args[2:], cli_args)
    elif cmd == 'kill':
        _kill(script_args[2:], cli_args)

    else:
        get_logger().critical('Unknown command ' + cmd)
        parser.print_help()
        print_help()
        return


def _list(args, cli_args):
    with model.get_db_provider(cli_args.config) as db:
        if len(args) == 0:
            experiments = db.get_user_experiments()
        elif args[0] == 'project':
            assert len(args) == 2
            experiments = db.get_project_experiments(args[1])
        elif args[0] == 'users':
            assert len(args) == 1
            users = db.get_users()
            for u in users:
                print(users[u].get('email'))
            return
        elif args[0] == 'user':
            assert len(args) == 2
            users = db.get_users()
            user_ids = [u for u in users if users[u].get('email') == args[1]]
            assert len(user_ids) == 1, \
                'The user with email ' + args[1] + \
                'not found!'
            experiments = db.get_user_experiments(user_ids[0])
        elif args[0] == 'all':
            assert len(args) == 1
            users = db.get_users()
            experiments = []
            for u in users:
                experiments += db.get_user_experiments(u)
        else:
            get_logger().critical('Unknown command ' + args[0])
            return

        if cli_args.short:
            for e in experiments:
                print(e)
            return

        experiments = [db.get_experiment(e) for e in experiments]

    experiments.sort(key=lambda e: -e.time_added)
    table = [['Time added', 'Key', 'Project', 'Status']]

    for e in experiments:
        table.append([
            time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(e.time_added)),
            e.key,
            e.project,
            e.status])

    print(AsciiTable(table).table)


def _stop(args, cli_args):
    with model.get_db_provider(cli_args.config) as db:
        for e in args:
            get_logger().info('Stopping experiment ' + e)
            db.stop_experiment(e)


def _kill(args, cli_args):
    with model.get_db_provider(cli_args.config) as db:
        for e in args:
            get_logger().info('Deleting experiment ' + e)
            db.delete_experiment(e)


def get_logger():
    global _my_logger
    if not _my_logger:
        _my_logger = logs.getLogger('studio-runs')
    return _my_logger


if __name__ == '__main__':
    main()



from keras.layers import merge
from keras.layers.core import Lambda
from keras.models import Model

import tensorflow as tf


def make_parallel(model, gpu_count):
    def get_slice(data, idx, parts):
        shape = tf.shape(data)
        size = tf.concat([shape[:1] // parts, shape[1:]], 0)
        stride = tf.concat([shape[:1] // parts, shape[1:] * 0], 0)
        start = stride * idx
        return tf.slice(data, start, size)

    outputs_all = []
    for i in range(len(model.outputs)):
        outputs_all.append([])

    # Place a copy of the model on each GPU, each getting a slice of the batch
    for i in range(gpu_count):
        with tf.device('/gpu:%d' % i):
            with tf.name_scope('tower_%d' % i):

                inputs = []
                # Slice each input into a piece for processing on this GPU
                for x in model.inputs:
                    input_shape = tuple(x.get_shape().as_list())[1:]
                    slice_n = Lambda(
                        get_slice, output_shape=input_shape, arguments={
                            'idx': i, 'parts': gpu_count})(x)
                    inputs.append(slice_n)

                outputs = model(inputs)

                if not isinstance(outputs, list):
                    outputs = [outputs]

                # Save all the outputs for merging back together later
                for l in range(len(outputs)):
                    outputs_all[l].append(outputs[l])

    # merge outputs on CPU
    with tf.device('/cpu:0'):
        merged = []
        for outputs in outputs_all:
            merged.append(merge(outputs, mode='concat', concat_axis=0))

        return Model(input=model.inputs, output=merged)



import os
import pickle
from studio import fs_tracker


def clientFunction(args, files):
    print('client function call with args ' +
          str(args) + ' and files ' + str(files))

    modelfile = 'model.dat'
    filename = files.get('model') or \
        os.path.join(fs_tracker.get_artifact('modeldir'), modelfile)

    print("Trying to load file {}".format(filename))

    if os.path.exists(filename):
        with open(filename, 'rb') as f:
            args = pickle.loads(f.read()) + 1

    else:
        print("Trying to write file {}".format(filename))
        with open(filename, 'wb') as f:
            f.write(pickle.dumps(args, protocol=2))

    return args


if __name__ == "__main__":
    clientFunction('test', {})



import os
import getpass
import time
import json
import shutil
import atexit
import tempfile
import requests
import re
import uuid

from builtins import input

try:
    from apscheduler.schedulers.background import BackgroundScheduler
except BaseException:
    BackgroundScheduler = None

import google.auth.transport.requests
import google.oauth2.id_token

from .util import rand_string
from . import pyrebase
from . import logs


TOKEN_DIR = os.path.expanduser('~/.studioml/keys')

HOUR = 3600
HALF_HOUR = 1800
API_KEY_COOLDOWN = 900
SLEEP_TIME = 0.5
MAX_NUM_RETRIES = 100


_auth_singleton = None

_grequest = google.auth.transport.requests.Request()


def get_auth_class(authtype):
    if authtype is None:
        return None
    elif authtype.lower() == 'none':
        return None
    elif authtype.lower() == 'firebase':
        return FirebaseAuth
    elif authtype.lower() == 'github':
        return GithubAuth
    else:
        raise ValueError(
            'Unknown authentication type {}, valid values are none, ' +
            'firebase, github'
            .format(authtype))


def get_auth(
        config,
        blocking=True,
        verbose=logs.DEBUG):

    global _auth_singleton
    if _auth_singleton is None:
        if isinstance(config, dict):
            auth_class = get_auth_class(config['type'])
        else:
            auth_class = get_auth_class(config)

        if auth_class:
            _auth_singleton = auth_class(
                config,
                blocking=blocking,
                verbose=verbose
            )
        else:
            return None
    return _auth_singleton


def get_and_verify_user(request, authtype):
    if not request.headers or 'Authorization' not in request.headers.keys():
        return None

    token = request.headers['Authorization'].split(' ')[-1]
    if not token or token == 'null':
        return None

    authclass = get_auth_class(authtype)
    if not authclass:
        return None

    if request.json:
        refresh_token = request.json.get('refreshToken')
    else:
        refresh_token = None

    return authclass.verify_token(token, refresh_token)


class GithubAuth(object):

    def __init__(
        self,
        config,
        blocking=True,
        verbose=logs.DEBUG
    ):
        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(verbose)

        if isinstance(config, dict):
            self.config = config
        else:
            self.config = {'type': config}

        self.tokendir = os.path.abspath(os.path.expanduser(
            self.config.get('token_directory', TOKEN_DIR))
        )

        if not os.path.exists(self.tokendir):
            os.makedirs(self.tokendir)

        self.token = self._load_token()[0]
        if self.token is None and blocking:
            self._sign_in()

    def get_user_id(self):
        return self.userid

    def get_token(self):
        if self.token is None:
            self.token = self._load_token()[0]

        return self.token

    def get_user_email(self):
        return self.userid

    def refresh_token(self, userid, refreshtoken):
        pass

    def is_expired(self):
        return False

    def _load_token(self):
        tokendir_contents = os.listdir(self.tokendir)
        tokens = [
            f for f in tokendir_contents
            if os.path.isfile(os.path.join(self.tokendir, f)) and
            f.endswith('.githubtoken')
        ]

        # TODO selection out of multiple token files
        if not any(tokens):
            return None, None

        token_file = os.path.join(self.tokendir, tokens[0])
        with open(token_file) as f:
            token = f.read().strip()

        self.userid = self.verify_token(token)
        if self.userid != re.sub(r'.githubtoken\Z', '', tokens[0]):
            self.logger.error(
                'Token filename does not match github login'
            )
            return None, None

        return token, token_file

    def get_token_file(self):
        return self._load_token()[1]

    def _save_token(self):
        if self.token is None:
            return

        token_file = os.path.join(self.tokendir, self.userid + '.githubtoken')
        with open(token_file, 'w') as f:
            f.write(self.token)

    def _sign_in(self):
        print(
            '*** \n' +
            'Sign in with your GitHub username and password \n' +
            'NOTE: studio.ml does NOT store GitHub passwords ' +
            ' or any other passwords for that matter. \n ' +
            'Your ' +
            'password will be used a single time to generate ' +
            'an authorization token that allows studio to '
            'verify your identity in the future.'
        )

        while True:
            self.userid = input('user id:')
            password = getpass.getpass('password:')

            response = requests.post(
                'https://api.github.com/authorizations',
                json={
                    'scopes': ['read:user'],
                    'note': 'studioml authorization, ' +
                            'random = ' + str(uuid.uuid4())
                },
                auth=(self.userid, password),
            )

            password = None

            if response.status_code == 201:
                self.token = response.json()['token']
                self._save_token()
                print('Successfully created token with name: ' +
                      response.json()['app']['name'])
                return

            else:
                print("GitHub login failure")
                print(response.json())

    @staticmethod
    def verify_token(token, refresh_token=None):
        response = requests.get(
            'https://api.github.com/user',
            headers={"Authorization": "Bearer " + token}
        )

        if response.status_code != 200:
            return None
        else:
            return response.json()['login']


class FirebaseAuth(object):
    def __init__(
            self,
            config,
            blocking=True,
            verbose=logs.DEBUG):
        if not os.path.exists(TOKEN_DIR):
            try:
                os.makedirs(TOKEN_DIR)
            except OSError:
                pass

        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(logs.DEBUG)

        self.firebase = pyrebase.initialize_app(config)
        self.user = {}
        self.use_email_auth = config.get('use_email_auth', False)
        if self.use_email_auth:
            self.email = config.get('email')
            self.password = config.get('password')
            if not self.password or not self.email:
                self.email = input(
                    'Firebase token is not found or expired! ' +
                    'You need to re-login. (Or re-run with ' +
                    'studio/studio-runner ' +
                    'with --guest option ) '
                    '\nemail:')
                self.password = getpass.getpass('password:')

        self.expired = True
        self.token_file = os.path.join(TOKEN_DIR, self.firebase.api_key)
        self._update_user()

        if self.expired and blocking:
            print('Authentication required! Either specify ' +
                  'use_email_auth in config file, or run '
                  'studio and go to webui ' +
                  '(localhost:5000 by default) '
                  'to authenticate using google credentials')
            while self.expired:
                time.sleep(1)
                self._update_user()

        self.sched = BackgroundScheduler()
        self.sched.start()
        self.sched.add_job(self._update_user, 'interval', minutes=31)
        atexit.register(self.sched.shutdown)

    def _update_user(self):
        if not os.path.exists(self.token_file):
            # refresh tokens don't expire, hence we can
            # use them forever once obtained
            # or (time.time() - os.path.getmtime(api_key)) > HOUR:
            if self.use_email_auth:
                self.sign_in_with_email()
                self.expired = False
            else:
                self.expired = True
        else:
            # If json file fails to load, try again
            counter = 0
            user = None
            while True:
                if user is not None or counter >= MAX_NUM_RETRIES:
                    break
                try:
                    with open(self.token_file) as f:
                        user = json.loads(f.read())
                except BaseException as e:
                    self.logger.info(e)
                    time.sleep(SLEEP_TIME)
                    counter += 1
            if user is None:
                return

            self.user = user
            if time.time() > self.user.get('expiration', 0):
                counter = 0
                while counter < MAX_NUM_RETRIES:
                    try:
                        self.refresh_token(user['email'], user['refreshToken'])
                        break
                    except BaseException as e:
                        self.logger.info(e)
                        time.sleep(SLEEP_TIME)
                        counter += 1
            else:
                self.expired = False

    def sign_in_with_email(self):
        self.user = \
            self.firebase.auth().sign_in_with_email_and_password(
                self.email,
                self.password)

        # TODO check if credentials worked

        self.user['email'] = self.email

    def refresh_token(self, email, refresh_token):
        self.user = self.firebase.auth().refresh(refresh_token)
        self.user['email'] = email
        self.user['expiration'] = time.time() + API_KEY_COOLDOWN
        self.expired = False

        # if not os.path.exists(api_key) or \
        #   time.time() - os.path.getmtime(api_key) > HALF_HOUR:
        # Rename to ensure atomic writes to json file
        # (technically more safe, but slower)

        tmp_api_key = os.path.join(tempfile.gettempdir(),
                                   "api_key_%s" % rand_string(32))
        with open(tmp_api_key, 'w') as f:
            f.write(json.dumps(self.user))
            f.flush()
            os.fsync(f.fileno())
            f.close()
        os.rename(tmp_api_key, self.token_file)

    def get_token(self):
        if self.expired:
            return None
        return self.user['idToken']

    def get_token_file(self):
        return self.token_file

    def get_user_id(self):
        if self.expired:
            return None

        if 'localId' in self.user.keys():
            return self.user['localId']

        return self.user['userId']

    def get_user_email(self):
        if self.expired:
            return None
        # we could also use the get_account_info
        # print self.firebase.auth().get_account_info(self.get_token())
        return self.user['email']

    def is_expired(self):
        return self.expired

    @staticmethod
    def verify_token(token, refresh_token=None):
        claims = google.oauth2.id_token.verify_firebase_token(
            token, _grequest)

        if not claims:
            return None
        else:
            if refresh_token:
                self.refresh_token(claims['email'], refresh_token)

            # get_db().register_user(claims['user_id'], claims['email'])

            return claims['user_id']


def remove_all_keys():
    keypath = os.path.join(os.path.expanduser('~'), '.studioml', 'keys')
    if os.path.exists(keypath):
        try:
            shutil.rmtree(keypath)
        except OSError:
            pass



import yaml

AWSInstance = "aws" in list(
    yaml.load(
        open(
            "tests/test_config.yaml",
            "r"))["cloud"].keys())
GcloudInstance = "gcloud" in list(
    yaml.load(
        open(
            "tests/test_config.yaml",
            "r"))["cloud"].keys())


def on_gcp():
    return GcloudInstance


def on_aws():
    return AWSInstance



import os
import re
import glob
import uuid
import sys
import time

try:
    try:
        from pip._internal.operations import freeze
    except Exception:
        from pip.operations import freeze
except ImportError:
    freeze = None

from . import fs_tracker
from .util import shquote


class Experiment(object):
    """Experiment information."""

    def __init__(self, key, filename, args, pythonenv,
                 project=None,
                 artifacts=None,
                 status='waiting',
                 resources_needed=None,
                 time_added=None,
                 time_started=None,
                 time_last_checkpoint=None,
                 time_finished=None,
                 info={},
                 git=None,
                 metric=None,
                 pythonver=None,
                 max_duration=None):

        self.key = key
        self.args = []
        self.filename = filename

        if filename and '::' in filename:
            self.filename = '-m'
            module_name = filename.replace('::', '.')
            if module_name.startswith('.'):
                module_name = module_name[1:]

            self.args.append(module_name)

        if args:
            self.args += args

        self.args = [shquote(a) for a in self.args]

        self.pythonenv = pythonenv
        self.project = project
        self.pythonver = pythonver if pythonver else str(sys.version_info[0]) + '.' + str(sys.version_info[1])

        workspace_path = os.path.abspath('.')
        try:
            model_dir = fs_tracker.get_model_directory(key)
        except BaseException:
            model_dir = None

        self.artifacts = {
            'workspace': {
                'local': workspace_path,
                'mutable': False,
                'unpack': True
            },
            'modeldir': {
                'local': model_dir,
                'mutable': True,
                'unpack': True
            },
            'output': {
                'local': fs_tracker.get_artifact_cache('output', key),
                'mutable': True,
                'unpack': True
            },
            'tb': {
                'local': fs_tracker.get_tensorboard_dir(key),
                'mutable': True,
                'unpack': True
            },
            '_metrics': {
                'local': fs_tracker.get_artifact_cache('_metrics', key),
                'mutable': True,
                'unpack': True
            },
            '_metadata': {
                'local': fs_tracker.get_artifact_cache('_metadata', key),
                'mutable': True,
                'unpack': True
            }
        }
        if artifacts is not None:
            self.artifacts.update(artifacts)

        self.resources_needed = resources_needed
        self.status = status
        self.time_added = time_added
        self.time_started = time_started
        self.time_last_checkpoint = time_last_checkpoint
        self.time_finished = time_finished
        self.info = info
        self.git = git
        self.metric = metric
        self.max_duration = max_duration

    def get_model(self, db):
        modeldir = db.get_artifact(self.artifacts['modeldir'])
        hdf5_files = [
            (p, os.path.getmtime(p))
            for p in
            glob.glob(modeldir + '/*.hdf*') +
            glob.glob(modeldir + '/*.h5')]
        if any(hdf5_files):
            # experiment type - keras
            import keras
            last_checkpoint = max(hdf5_files, key=lambda t: t[1])[0]
            return keras.models.load_model(last_checkpoint)

        if self.info.get('type') == 'tensorflow':
            raise NotImplementedError

        raise ValueError("Experiment type is unknown!")


def create_experiment(
        filename,
        args,
        experiment_name=None,
        project=None,
        artifacts={},
        resources_needed=None,
        metric=None,
        max_duration=None):
    key = experiment_name if experiment_name else \
        str(int(time.time())) + "_" + str(uuid.uuid4())

    packages = []
    for pkg in freeze.freeze():

        if pkg.startswith('-e git+'):
            # git package
            packages.append(pkg)
        elif '==' in pkg:
            # pypi package
            pkey = re.search(r'^.*?(?=\=\=)', pkg).group(0)
            pversion = re.search(r'(?<=\=\=).*\Z', pkg).group(0)

            if resources_needed is not None and \
                    int(resources_needed.get('gpus')) > 0:
                if (pkey == 'tensorflow' or key == 'tf-nightly'):
                    pkey = pkey + '-gpu'

            # TODO add installation logic for torch
            packages.append(pkey + '==' + pversion)

    return Experiment(
        key=key,
        filename=filename,
        args=args,
        pythonenv=[p for p in packages],
        project=project,
        artifacts=artifacts,
        resources_needed=resources_needed,
        metric=metric,
        max_duration=max_duration)


def experiment_from_dict(data, info={}):
    try:
        return Experiment(
            key=data['key'],
            filename=data['filename'],
            args=data.get('args'),
            pythonenv=data['pythonenv'],
            project=data.get('project'),
            status=data['status'],
            artifacts=data.get('artifacts'),
            resources_needed=data.get('resources_needed'),
            time_added=data['time_added'],
            time_started=data.get('time_started'),
            time_last_checkpoint=data.get('time_last_checkpoint'),
            time_finished=data.get('time_finished'),
            info=info if any(info) else data.get('info'),
            git=data.get('git'),
            metric=data.get('metric'),
            pythonver=data.get('pythonver'),
            max_duration=data.get('max_duration')
        )
    except KeyError as e:
        self.logger.error(data)
        raise e



import tensorflow as tf
import sys
import os
import time

from keras.objectives import categorical_crossentropy
from keras.layers import Dense
from tensorflow.examples.tutorials.mnist import input_data

from keras import backend as K
from studio import fs_tracker

import logging

logging.basicConfig()

sess = tf.Session()
K.set_session(sess)
# Now let's get started with our MNIST model. We can start building a
# classifier exactly as you would do in TensorFlow:

# this placeholder will contain our input digits, as flat vectors
img = tf.placeholder(tf.float32, shape=(None, 784))
# We can then use Keras layers to speed up the model definition process:


# Keras layers can be called on TensorFlow tensors:
# fully-connected layer with 128 units and ReLU activation
x = Dense(128, activation='relu')(img)
x = Dense(128, activation='relu')(x)
# output layer with 10 units and a softmax activation
preds = Dense(10, activation='softmax')(x)
# We define the placeholder for the labels, and the loss function we will use:

labels = tf.placeholder(tf.float32, shape=(None, 10))

loss = tf.reduce_mean(categorical_crossentropy(labels, preds))
# Let's train the model with a TensorFlow optimizer:

mnist_data = input_data.read_data_sets('MNIST_data', one_hot=True)

global_step = tf.Variable(0, name='global_step', trainable=False)
train_step = tf.train.GradientDescentOptimizer(
    0.5).minimize(loss, global_step=global_step)
# Initialize all variables
init_op = tf.global_variables_initializer()
saver = tf.train.Saver()
sess.run(init_op)


logger = logging.getLogger('train_mnist')
logger.setLevel(10)
# Run training loop
with sess.as_default():
    while True:
        batch = mnist_data.train.next_batch(50)
        train_step.run(feed_dict={img: batch[0],
                                  labels: batch[1]})

        sys.stdout.flush()
        saver.save(
            sess,
            os.path.join(
                fs_tracker.get_model_directory(),
                "ckpt"),
            global_step=global_step)
        time.sleep(1)



import unittest
import inspect
import yaml
import uuid
import os
import time
import six
import shutil

from studio import model
from studio.firebase_provider import FirebaseProvider
from studio.postgres_provider import PostgresProvider
from studio.s3_provider import S3Provider
from studio.auth import remove_all_keys
from studio.util import has_aws_credentials
from env_detect import on_gcp, on_aws

from model_test import get_test_experiment


def get_methods(cls):
    methods = inspect.getmembers(cls, predicate=inspect.ismethod)
    return set([name for name, _ in methods if not name.startswith('_')])


class ProvidersTest(unittest.TestCase):

    def test_providers_compatible(self):
        # Check that all available providers are compatible.
        firebase_methods = get_methods(FirebaseProvider)
        postgres_methods = get_methods(PostgresProvider)
        s3_methods = get_methods(S3Provider)

        self.assertEqual(firebase_methods, postgres_methods)
        self.assertEqual(firebase_methods, s3_methods)


class KeyValueProviderTest(object):
    def get_default_config_name(self):
        return 'test_config.yaml'

    def get_provider(self, config_name=None):
        config_name = config_name if config_name else \
            self.get_default_config_name()

        config_file = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)),
            config_name)
        with open(config_file) as f:
            config = yaml.load(f)

        return model.get_db_provider(config)

    def test_get_set(self):
        with self.get_provider() as fb:
            fb._set("test/hello", "world")
            response = fb._get("test/hello")
            self.assertEquals(response, "world")

            random_str = str(uuid.uuid4())
            key_path = 'test/randomKey'
            fb._set(key_path, random_str)

            self.assertTrue(fb._get(key_path) == random_str)
            fb._delete(key_path)

    def test_get_user_keybase(self):
        with self.get_provider() as fb:
            keybase = fb._get_user_keybase()
            self.assertTrue(keybase == 'users/guest/')

    def test_get_experiments_keybase(self):
        with self.get_provider() as fb:
            keybase = fb._get_experiments_keybase()
            self.assertTrue(keybase == 'experiments/')

    def test_get_projects_keybase(self):
        with self.get_provider() as fb:
            keybase = fb._get_projects_keybase()
            self.assertTrue(keybase == 'projects/')

    def test_add_experiment(self):
        with self.get_provider() as fb:
            experiment, experiment_name, _, _ = get_test_experiment()

            fb._delete(fb._get_experiments_keybase() + experiment_name)
            fb.add_experiment(experiment)

            self.assertTrue(experiment.status == 'waiting')
            self.assertTrue(experiment.time_added <= time.time())
            actual_experiment_dict = fb._get(
                fb._get_experiments_keybase() + experiment_name)
            for key, value in six.iteritems(experiment.__dict__):
                if value:
                    self.assertTrue(actual_experiment_dict[key] == value)

            fb.finish_experiment(experiment)
            fb.delete_experiment(experiment)

    def test_start_experiment(self):
        with self.get_provider() as fb:
            experiment, experiment_name, _, _ = get_test_experiment()

            fb._delete(fb._get_experiments_keybase() + experiment_name)
            fb.add_experiment(experiment)
            fb.start_experiment(experiment)

            self.assertTrue(experiment.status == 'running')
            self.assertTrue(experiment.time_added <= time.time())
            self.assertTrue(experiment.time_started <= time.time())

            actual_experiment_dict = fb._get(
                fb._get_experiments_keybase() + experiment_name)
            for key, value in six.iteritems(experiment.__dict__):
                if value:
                    self.assertTrue(actual_experiment_dict[key] == value)

            fb.finish_experiment(experiment)
            fb.delete_experiment(experiment)

    def test_finish_experiment(self):
        with self.get_provider() as fb:
            experiment, experiment_name, _, _ = get_test_experiment()

            fb._delete(fb._get_experiments_keybase() + experiment_name)
            fb.add_experiment(experiment)
            fb.start_experiment(experiment)
            fb.finish_experiment(experiment)

            self.assertTrue(experiment.status == 'finished')
            self.assertTrue(experiment.time_added <= time.time())
            self.assertTrue(experiment.time_started <= time.time())
            self.assertTrue(experiment.time_finished <= time.time())

            actual_experiment_dict = fb._get(
                fb._get_experiments_keybase() + experiment_name)
            for key, value in six.iteritems(experiment.__dict__):
                if value:
                    self.assertTrue(actual_experiment_dict[key] == value)

            fb.delete_experiment(experiment)

    def test_checkpoint_experiment(self):
        with self.get_provider() as fb:
            experiment, experiment_name, _, _, = get_test_experiment()

            modeldir = experiment.artifacts['modeldir']['local']
            if os.path.exists(modeldir):
                shutil.rmtree(modeldir)

            os.makedirs(modeldir)
            try:
                fb.delete_experiment(experiment_name)
            except BaseException:
                pass

            fb.add_experiment(experiment)
            fb.start_experiment(experiment)

            file_in_modeldir = os.path.join(modeldir, str(uuid.uuid4()))
            random_str = str(uuid.uuid4())
            with open(file_in_modeldir, 'w') as f:
                f.write(random_str)

            checkpoint_threads = fb.checkpoint_experiment(experiment)
            if checkpoint_threads:
                for t in checkpoint_threads:
                    t.join()

            shutil.rmtree(modeldir)
            fb.store.get_artifact(
                fb.get_experiment(
                    experiment_name,
                    getinfo=False).artifacts['modeldir'])

            with open(file_in_modeldir, 'r') as f:
                line = f.read()

            self.assertTrue(line == random_str)
            fb.delete_experiment(experiment)


class FirebaseProviderTest(unittest.TestCase, KeyValueProviderTest):
    _multiprocess_shared_ = True

    def test_get_set_auth_firebase(self):
        remove_all_keys()
        with self.get_provider('test_config_auth.yaml') as fb:
            response = fb._get("authtest/hello")
            self.assertEquals(response, "world")

            random_str = str(uuid.uuid4())
            key_path = 'authtest/randomKey'
            fb._set(key_path, random_str)

            self.assertTrue(fb._get(key_path) == random_str)
            fb._delete(key_path)
            remove_all_keys()

    def test_get_set_noauth_firebase(self):
        remove_all_keys()
        with self.get_provider('test_config.yaml') as fb:
            response = fb._get("authtest/hello")
            self.assertTrue(response is None)

            random_str = str(uuid.uuid4())
            key_path = 'authtest/randomKey'
            fb._set(key_path, random_str)
            self.assertTrue(fb._get(key_path) is None)
            remove_all_keys()

    def test_get_set_firebase_bad(self):
        # smoke test to make sure access to a database at wrong
        # url is reported, but does not crash the system
        with self.get_provider('test_bad_config.yaml') as fb:
            response = fb._get("test/hello")
            self.assertTrue(response is None)

            fb._set("test/hello", "bla")


@unittest.skipIf(
    not on_aws(),
    'User indicated not on aws')
class UserIndicatedOnAWSTest(unittest.TestCase):
    def test_on_enviornment(self):
        self.assertTrue(has_aws_credentials())


@unittest.skipIf(
    (not on_aws()) or not has_aws_credentials(),
    'Skipping due to userinput or AWS Not detected')
class S3ProviderTest(unittest.TestCase, KeyValueProviderTest):
    _multiprocess_shared_ = True

    def get_default_config_name(self):
        return 'test_config_s3.yaml'


@unittest.skipIf(
    'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
    'google app credentials not found, cannot run test')
class GSProviderTest(unittest.TestCase, KeyValueProviderTest):
    def get_default_config_name(self):
        return 'test_config_gs.yaml'


if __name__ == "__main__":
    unittest.main()



import os
from . import fs_tracker, logs
import uuid
import glob
import time
import filelock

_local_queue_lock = filelock.FileLock(
    os.path.expanduser('~/.studioml/local_queue.lock')
)


class LocalQueue:
    def __init__(self, path=None, verbose=10):
        if path is None:
            self.path = fs_tracker.get_queue_directory()
        else:
            self.path = path
        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(verbose)

    def has_next(self):
        return any(glob.glob(self.path + '/*'))

    def clean(self, timeout=0):
        while self.has_next():
            self.dequeue()

    # Delete and clean are the same for local queue
    def delete(self):
        self.clean()

    def dequeue(self, acknowledge=True, timeout=0):
        wait_step = 1
        for waited in range(0, timeout + wait_step, wait_step):
            with _local_queue_lock:
                files = glob.glob(self.path + '/*')
                if any(files):
                    first_file = min([(p, os.path.getmtime(p)) for p in files],
                                     key=lambda t: t[1])[0]

                    with open(first_file, 'r') as f:
                        data = f.read()

                    self.acknowledge(first_file)
                    if not acknowledge:
                        return data, first_file
                    else:
                        return data

                elif waited == timeout:
                    return None

            # self.logger.info(
            #    ('No messages found, sleeping for {} ' +
            #     ' (total sleep time {})').format(wait_step, waited))
            time.sleep(wait_step)

    def enqueue(self, data):
        with _local_queue_lock:
            filename = os.path.join(self.path, str(uuid.uuid4()))
            with open(filename, 'w') as f:
                f.write(data)

    def acknowledge(self, key):
        try:
            os.remove(key)
        except BaseException:
            pass

    def hold(self, key, minutes):
        self.acknowledge(key)

    def get_name(self):
        return 'local'


def get_local_queue_lock():
    return _local_queue_lock



import os
import numpy as np

from keras.layers import Dense
from keras.models import Sequential

from studio import fs_tracker

model = Sequential()
model.add(Dense(2, input_shape=(2,)))

weights = model.get_weights()
new_weights = [np.array([[2, 0], [0, 2]])]
# print weights
# new_weights = []
# for weight in weights:
#   new_weights.append(weight + 1)

model.set_weights(new_weights)
model.save(os.path.join(fs_tracker.get_model_directory(), 'weights.h5'))



import sys
from keras.layers import Input, Dense
from keras.models import Model
from keras.datasets import mnist
from keras.utils import to_categorical
from keras.callbacks import ModelCheckpoint, TensorBoard

from studio import fs_tracker
from studio.multi_gpu import make_parallel

# this placeholder will contain our input digits, as flat vectors
img = Input((784,))
# fully-connected layer with 128 units and ReLU activation
x = Dense(128, activation='relu')(img)
x = Dense(128, activation='relu')(x)
# output layer with 10 units and a softmax activation
preds = Dense(10, activation='softmax')(x)


no_gpus = 2
batch_size = 128

model = Model(img, preds)
model = make_parallel(model, no_gpus)
model.compile(loss='categorical_crossentropy', optimizer='adam')

(x_train, y_train), (x_test, y_test) = mnist.load_data()

x_train = x_train.reshape(60000, 784)
x_test = x_test.reshape(10000, 784)
x_train = x_train.astype('float32')
x_test = x_test.astype('float32')
x_train /= 255
x_test /= 255

# convert class vectors to binary class matrices
y_train = to_categorical(y_train, 10)
y_test = to_categorical(y_test, 10)


checkpointer = ModelCheckpoint(
    fs_tracker.get_model_directory() +
    '/checkpoint.{epoch:02d}-{val_loss:.2f}.hdf')


tbcallback = TensorBoard(log_dir=fs_tracker.get_tensorboard_dir(),
                         histogram_freq=0,
                         write_graph=True,
                         write_images=False)


model.fit(
    x_train,
    y_train,
    validation_data=(x_test, y_test),
    epochs=int(sys.argv[1]),
    batch_size=batch_size * no_gpus,
    callbacks=[checkpointer, tbcallback])



import os
import json

from .keyvalue_provider import KeyValueProvider
from .local_artifact_store import LocalArtifactStore

class LocalDbProvider(KeyValueProvider):

    def __init__(self, config, blocking_auth=True, verbose=10, store=None):
        self.config = config
        self.bucket = config.get('bucket', 'studioml-meta')

        self.endpoint = config.get('endpoint', '~')
        self.db_root = os.path.realpath(os.path.expanduser(self.endpoint))
        if not os.path.exists(self.db_root) \
            or not os.path.isdir(self.db_root):
            raise ValueError("Local DB root {} doesn't exist or not a directory!".format(self.db_root))

        self.bucket = config.get('bucket')
        self.db_root = os.path.join(self.db_root, self.bucket)
        self._ensure_path_dirs_exist(self.db_root)

        super(LocalDbProvider, self).__init__(
            config,
            blocking_auth,
            verbose,
            store)

    def _ensure_path_dirs_exist(self, path):
        dirs = os.path.dirname(path)
        os.makedirs(dirs, mode = 0o777, exist_ok = True)

    def _get(self, key, shallow=False):
        file_name = os.path.join(self.db_root, key)
        if not os.path.exists(file_name):
            return None
        with open(file_name) as infile:
            result = json.load(infile)
        return result

    def _delete(self, key):
        file_name = os.path.join(self.db_root, key)
        if os.path.exists(file_name):
            os.remove(file_name)

    def _set(self, key, value):
        file_name = os.path.join(self.db_root, key)
        self._ensure_path_dirs_exist(file_name)
        with open(file_name, 'w') as outfile:
            json.dump(value, outfile)
        



import unittest
import numpy as np
import six

from PIL import Image
from io import BytesIO

try:
    import keras
    from keras.layers import Dense, Flatten
    from keras.models import Sequential
    from keras.datasets import mnist
    from keras.utils import to_categorical

except ImportError:
    keras = None

from timeout_decorator import timeout

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty
from studio import model_util


class ModelUtilTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def test_q2q_batch(self):
        data = six.moves.range(10)
        q_in = Queue()
        q_out = Queue()

        for d in data:
            q_in.put(d)

        model_util._q2q_batch(
            lambda b: [
                x * x for x in b],
            q_in,
            q_out,
            batch_size=4)

        expected_out = [x * x for x in data]
        actual_out = []
        while True:
            try:
                actual_out.append(q_out.get(True, 1))
            except Empty:
                break

        self.assertEquals(expected_out, actual_out)

    def test_q2q_batch_filter(self):
        data = six.moves.range(10)
        q_in = Queue()
        q_out = Queue()

        for d in data:
            q_in.put(d)

        model_util._q2q_batch(
            lambda b: [
                x * x for x in b],
            q_in,
            q_out,
            filterf=lambda x: x != 3,
            batch_size=4)

        expected_out = [x * x for x in data if x != 3]
        actual_out = []
        while True:
            try:
                actual_out.append(q_out.get(True, 1))
            except Empty:
                break

        self.assertEquals(expected_out, actual_out)

    def test_q2gen_gen2q(self):
        data = (x for x in range(10))
        q = Queue()

        model_util._gen2q(data, q)

        expected_out = list(range(10))
        actual_out = []
        while True:
            try:
                actual_out.append(q.get(True, 1))
            except Empty:
                break

        self.assertEquals(expected_out, actual_out)

        model_util._gen2q((x for x in range(10)), q)
        gen = model_util._q2gen(q)

        self.assertEquals(expected_out, list(gen))


class BufferedPipeTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def test_pipe_simple(self):
        p = model_util.BufferedPipe() \
            .add(lambda x: x + 1) \
            .add(lambda x: x * x)

        lst = list(p((x for x in range(1, 5))))

        expected_l = [4, 9, 16, 25]

        self.assertEquals(lst, expected_l)

    @unittest.skip('ordering problem - peterz to fix')
    def test_pipe_buffer(self):
        p = model_util.BufferedPipe() \
            .add(lambda x: x + 1, num_workers=32) \
            .add(lambda x: x * x)

        out_gen = p((x for x in range(1, 5)))

        expected_l = [4, 9, 16, 25]

        self.assertEquals(list(out_gen), expected_l)


class ModelPipeTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def test_model_pipe(self):

        p = model_util.ModelPipe()
        p.add(lambda x: x * x, num_workers=32)

        input_dict = {x: x for x in range(10)}
        output_dict = p(input_dict)

        expected_dict = {k: v * v for k, v in six.iteritems(input_dict)}

        self.assertEquals(expected_dict, output_dict)

    @unittest.skip('peterz fix - fails in python3.6')
    @timeout(30)
    def test_model_pipe_long(self):

        p = model_util.ModelPipe()
        p.add(lambda x: x * x, num_workers=32, timeout=0.5)

        input_dict = {x: x for x in range(10000)}
        output_dict = p(input_dict)

        expected_dict = {k: v * v for k, v in six.iteritems(input_dict)}

        self.assertEquals(expected_dict, output_dict)

    def test_model_pipe_ordered(self):

        p = model_util.ModelPipe()
        p.add(lambda x: x * x, num_workers=32)

        input_list = list(range(10))
        output_list = p.apply_ordered(input_list)

        expected_list = [x * x for x in input_list]

        self.assertEquals(expected_list, output_list)

    def test_model_pipe_batch(self):

        p = model_util.ModelPipe()
        p.add(lambda b: [x * x for x in b], batch_size=4)

        input_list = list(range(10))
        output_list = p.apply_ordered(input_list)

        expected_list = [x * x for x in input_list]

        self.assertEquals(expected_list, output_list)

    def test_model_pipe_exception(self):

        p = model_util.ModelPipe()
        p.add(lambda x: 1.0 / x, num_workers=4)

        input_list = list(range(-10, 10))
        output_list = p.apply_ordered(input_list)

        expected_list = [1.0 / x if x != 0 else None for x in input_list]

        self.assertEquals(expected_list, output_list)


@unittest.skipIf(keras is None,
                 "These tests require keras")
class KerasModelPipeTest(unittest.TestCase):
    def test_model_pipe_keras(self):
        model = Sequential()
        model.add(Flatten(input_shape=(1, 28, 28)))
        model.add(Dense(128, activation='relu'))
        model.add(Dense(10, activation='softmax'))

        p = model_util.ModelPipe()
        input_data = [np.random.random((1, 1, 28, 28)) for _ in range(2)]

        p.add(model.predict, batch_size=64, batcher=np.vstack)

        expected_output = [
            model.predict(
                x.reshape(
                    (1, 1, 28, 28))) for x in input_data]
        output = p.apply_ordered(input_data)

        self.assertTrue(np.isclose(np.array(output).flatten(),
                                   np.array(expected_output).flatten()).all())

    def test_model_pipe_mnist_urls(self):

        (x_train, y_train), (x_test, y_test) = mnist.load_data()

        x_train = x_train.reshape(60000, 28, 28, 1)
        x_test = x_test.reshape(10000, 28, 28, 1)
        x_train = x_train.astype('float32')
        x_test = x_test.astype('float32')
        x_train /= 255
        x_test /= 255

        y_train = to_categorical(y_train, 10)
        y_test = to_categorical(y_test, 10)

        model = Sequential()

        model.add(Flatten(input_shape=(28, 28, 1)))
        model.add(Dense(128, activation='relu'))
        model.add(Dense(128, activation='relu'))

        model.add(Dense(10, activation='softmax'))

        no_epochs = 3
        batch_size = 32

        model.compile(loss='categorical_crossentropy', optimizer='adam')

        model.fit(
            x_train, y_train, validation_data=(
                x_test,
                y_test),
            epochs=no_epochs,
            batch_size=batch_size)

        pipe = model_util.ModelPipe()

        pipe.add(
            lambda url: six.moves.urllib.request.urlopen(url).read(),
            num_workers=2,
            timeout=10)
        pipe.add(lambda img: Image.open(BytesIO(img)))
        pipe.add(model_util.resize_to_model_input(model))
        pipe.add(lambda x: 1 - x)
        pipe.add(model, num_workers=1, batch_size=32, batcher=np.vstack)
        pipe.add(lambda x: np.argmax(x, axis=1))

        url5 = 'http://blog.otoro.net/assets/20160401/png/mnist_output_10.png'
        url2 = 'http://joshmontague.com/images/mnist-2.png'
        urlb = 'http://joshmontague.com/images/mnist-3.png'

        expected_output = {url5: 5, url2: 2}
        output = pipe({url5: url5, url2: url2, urlb: urlb})
        output = {k: v for k, v in six.iteritems(output) if v}

        self.assertEquals(output, expected_output)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from studio.hyperparameter import HyperparameterParser, Hyperparameter
from studio import logs


class RunnerArgs(object):
    def __init__(self):
        self.optimizer = "grid"
        self.verbose = False


class HyperparamTest(unittest.TestCase):
    def test_parse_range(self):
        logger = logs.getLogger('test_stop_experiment')
        h = HyperparameterParser(RunnerArgs(), logger)
        range_strs = ['1,2,3', ':5', '2:5', '0.1:0.05:0.3', '0.1:3:0.3',
                      '0.01:4l:10']
        gd_truths = [
            [
                1.0, 2.0, 3.0], [
                0.0, 1.0, 2.0, 3.0, 4.0, 5.0], [
                2.0, 3.0, 4.0, 5.0], [
                    0.1, 0.15, 0.2, 0.25, 0.3], [
                        0.1, 0.2, 0.3], [
                            0.01, 0.1, 1, 10]]

        for range_str, gd_truth in zip(range_strs, gd_truths):
            hyperparameter = h._parse_grid("test", range_str)
            self.assertTrue(np.isclose(hyperparameter.values, gd_truth).all())

    def test_unfold_tuples(self):
        logger = logs.getLogger('test_stop_experiment')
        h = HyperparameterParser(RunnerArgs(), logger)

        hyperparams = [Hyperparameter(name='a', values=[1, 2, 3]),
                       Hyperparameter(name='b', values=[4, 5])]

        expected_tuples = [
            {'a': 1, 'b': 4}, {'a': 2, 'b': 4}, {'a': 3, 'b': 4},
            {'a': 1, 'b': 5}, {'a': 2, 'b': 5}, {'a': 3, 'b': 5}]

        self.assertEquals(
            sorted(h.convert_to_tuples(hyperparams), key=lambda x: str(x)),
            sorted(expected_tuples, key=lambda x: str(x)))


if __name__ == '__main__':
    unittest.main()



import os
import subprocess
import xml.etree.ElementTree as ET

from .util import sixdecode


def memstr2int(string):
    conversion_factors = [
        ('Mb', 2**20), ('MiB', 2**20), ('m', 2**20), ('mb', 2**20),
        ('Gb', 2**30), ('GiB', 2**30), ('g', 2**30), ('gb', 2**30),
        ('kb', 2**10), ('k', 2**10)
    ]

    for k, f in conversion_factors:
        if string.endswith(k):
            return int(float(string.replace(k, '')) * f)

    return int(string)


def get_available_gpus(gpu_mem_needed=None, strict=False):
    gpus = _get_gpu_info()

    def check_gpu_nomem_strict(gpu):
        return memstr2int(gpu.find('fb_memory_usage').find('used').text) < \
            memstr2int(gpu.find('fb_memory_usage').find('free').text)
    if gpu_mem_needed is None:
        if strict:
            return [gpu.find('minor_number').text for gpu in gpus if
                    check_gpu_nomem_strict(gpu)]
        else:
            return [gpu.find('minor_number').text for gpu in gpus]

    gpu_mem_needed = memstr2int(gpu_mem_needed)

    def check_gpu_mem_strict(gpu):
        return gpu_mem_needed < \
            memstr2int(gpu.find('fb_memory_usage').find('free').text)

    def check_gpu_mem_loose(gpu):
        return (gpu_mem_needed <
                memstr2int(gpu.find('fb_memory_usage').find('total').text)) \
            and check_gpu_nomem_strict(gpu)

    if strict:
        return [gpu.find('minor_number').text
                for gpu in gpus if check_gpu_mem_strict(gpu)]
    else:
        return [gpu.find('minor_number').text
                for gpu in gpus if check_gpu_mem_loose(gpu)]


def _get_gpu_info():
    try:
        smi_proc = subprocess.Popen(['nvidia-smi', '-q', '-x'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)

        smi_output, _ = smi_proc.communicate()
        xmlroot = ET.fromstring(sixdecode(smi_output))

        return xmlroot.findall('gpu')
    except Exception:
        return []


def get_gpus_summary():
    info = _get_gpu_info()

    def info_to_summary(gpuinfo):
        util = gpuinfo.find('utilization').find('gpu_util').text
        mem = gpuinfo.find('fb_memory_usage').find('used').text

        return "util: {}, mem {}".format(util, memstr2int(mem))

    return " ".join([
        "gpu {} {}".format(
            gpuinfo.find('minor_number').text,
            info_to_summary(gpuinfo)) for gpuinfo in info])


def get_gpu_mapping():
    no_gpus = len(_get_gpu_info())
    return {str(i): i for i in range(no_gpus)}


def _find_my_gpus(prop='minor_number'):
    gpu_info = _get_gpu_info()
    my_gpus = [g.find(prop).text for g in gpu_info if os.getpid() in [int(
        p.find('pid').text) for p in
        g.find('processes').findall('process_info')]]

    return my_gpus


if __name__ == "__main__":
    print(get_gpu_mapping())



import calendar
import os
import shutil

from .tartifact_store import TartifactStore

class LocalArtifactStore(TartifactStore):
    def __init__(self, config,
                 bucket_name=None,
                 verbose=10,
                 measure_timestamp_diff=False,
                 compression=None):

        if compression is None:
            compression = config.get('compression')

        self.endpoint = config.get('endpoint', '~')
        self.store_root = os.path.realpath(os.path.expanduser(self.endpoint))
        if not os.path.exists(self.store_root) \
            or not os.path.isdir(self.store_root):
            raise ValueError()

        self.bucket = bucket_name
        if self.bucket is None:
            self.bucket = config.get('bucket')
        self.store_root = os.path.join(self.store_root, self.bucket)
        self._ensure_path_dirs_exist(self.store_root)

        super(LocalArtifactStore, self).__init__(
            measure_timestamp_diff,
            compression=compression,
            verbose=verbose)

    def _ensure_path_dirs_exist(self, path):
        dirs = os.path.dirname(path)
        os.makedirs(dirs, mode = 0o777, exist_ok = True)

    def _upload_file(self, key, local_path):
        target_path = os.path.join(self.store_root, key)
        self._ensure_path_dirs_exist(target_path)
        shutil.copyfile(local_path, target_path)

    def _download_file(self, key, local_path, bucket=None):
        source_path = os.path.join(self.store_root, key)
        self._ensure_path_dirs_exist(local_path)
        shutil.copyfile(source_path, local_path)

    def _delete_file(self, key):
        os.remove(os.path.join(self.store_root, key))

    def _get_file_url(self, key, method='GET'):
        return str(os.path.join(self.store_root, key))

    def _get_file_post(self, key):
        return str(os.path.join(self.store_root, key))

    def _get_file_timestamp(self, key):
            return None

    def get_qualified_location(self, key):
        return 'file:/' + self.store_root + '/' + key

    def get_bucket(self):
        return self.bucket



from .tartifact_store import TartifactStore
from .util import download_file, upload_file
from . import logs


class HTTPArtifactStore(TartifactStore):
    def __init__(self, url,
                 timestamp=None,
                 compression=None,
                 verbose=logs.DEBUG):

        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(verbose)

        self.url = url
        self.timestamp = timestamp

        super(HTTPArtifactStore, self).__init__(
            False,
            compression=compression,
            verbose=verbose)

    def _upload_file(self, key, local_path):
        upload_file(self.url, local_path, self.logger)

    def _download_file(self, key, local_path):
        download_file(self.url, local_path, self.logger)

    def _delete_file(self, key):
        raise NotImplementedError

    def _get_file_url(self, key):
        raise NotImplementedError

    def _get_file_timestamp(self, key):
        return self.timestamp

    def get_qualified_location(self, key):
        return self.endpoint + '/' + key



import re
import subprocess
import os


def get_git_info(path='.', abort_dirty=True):
    info = {}
    if not is_git(path):
        return None

    if abort_dirty and not is_clean(path):
        return None

    info['url'] = get_repo_url(path)
    info['commit'] = get_commit(path)
    return info


def is_git(path='.'):
    p = subprocess.Popen(
        ['git', 'status'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=path)

    p.wait()
    return (p.returncode == 0)


def is_clean(path='.'):
    p = subprocess.Popen(
        ['git', 'status', '-s'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=path)

    stdout, _ = p.communicate()
    if not p.returncode == 0:
        return False

    return (stdout.strip() == '')


def get_repo_url(path='.', remove_user=True):
    p = subprocess.Popen(
        ['git', 'config', '--get', 'remote.origin.url'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=path)

    stdout, _ = p.communicate()
    if p.returncode != 0:
        return None

    url = stdout.strip()
    if remove_user:
        url = re.sub('(?<=://).*@', '', url.decode('utf-8'))
    return url


def get_branch(path='.'):
    p = subprocess.Popen(
        ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=path)

    stdout, _ = p.communicate()
    if p.returncode == 0:
        return None

    return stdout.strip().decode('utf8')


def get_commit(path='.'):
    p = subprocess.Popen(
        ['git', 'rev-parse', 'HEAD'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=path)

    stdout, _ = p.communicate()
    if p.returncode != 0:
        return None

    return stdout.strip().decode('utf8')


def get_my_repo_url():
    mypath = os.path.dirname(os.path.realpath(__file__))
    repo = get_repo_url(mypath)
    if repo is None:
        repo = "https://github.com/studioml/studio"
    return repo


def get_my_branch():
    mypath = os.path.dirname(os.path.realpath(__file__))
    branch = get_branch(mypath)
    if branch is None:
        branch = "master"
    return branch


def get_my_checkout_target():
    mypath = os.path.dirname(os.path.realpath(__file__))
    target = get_commit(mypath)
    if target is None:
        target = get_my_branch()

    return target



import unittest
import subprocess
import time
from random import randint
import os
import uuid
import requests

from studio import model
from env_detect import on_gcp


@unittest.skipIf(
    not on_gcp(),
    'User indicated not on gcp')
class UserIndicatedOnGCPTest(unittest.TestCase):
    def test_on_enviornment(self):
        self.assertTrue('GOOGLE_APPLICATION_CREDENTIALS' in os.environ.keys())


@unittest.skipIf(
    (not on_gcp()) or
    'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
    'Skipping due to userinput or GCP Not detected' +
    "GOOGLE_APPLICATION_CREDENTIALS is missing, needed for " +
    "server to communicate with storage")
class ServingTest(unittest.TestCase):

    _mutliprocess_shared_ = True

    def _test_serving(self, data_in, expected_data_out, wrapper=None):

        self.port = randint(5000, 9000)
        server_experimentid = 'test_serving_' + str(uuid.uuid4())

        args = [
            'studio', 'run',
            '--force-git',
            '--verbose=debug',
            '--experiment=' + server_experimentid,
            '--config=' + self.get_config_path(),
            'studio::serve_main',
            '--port=' + str(self.port),
            '--host=localhost'
        ]

        if wrapper:
            args.append('--wrapper=' + wrapper)

        subprocess.Popen(args, cwd=os.path.dirname(__file__))
        time.sleep(60)

        try:
            retval = requests.post(
                url='http://localhost:' + str(self.port), json=data_in)
            data_out = retval.json()
            assert data_out == expected_data_out

        finally:
            with model.get_db_provider(model.get_config(
                    self.get_config_path())) as db:

                db.stop_experiment(server_experimentid)
                time.sleep(20)
                db.delete_experiment(server_experimentid)

    def test_serving_identity(self):
        data = {"a": "b"}
        self._test_serving(
            data_in=data,
            expected_data_out=data
        )

    def test_serving_increment(self):
        data_in = {"a": 1}
        data_out = {"a": 2}

        self._test_serving(
            data_in=data_in,
            expected_data_out=data_out,
            wrapper='model_increment.py'
        )

    def get_config_path(self):
        return os.path.join(
            os.path.dirname(__file__),
            'test_config_http_client.yaml'
        )


if __name__ == '__main__':
    unittest.main()



class BaseArtifactStore(object):

    def __init__(self):
        self.storage_client = None

    def set_storage_client(self, sclient):
        self.storage_client = sclient

    def get_storage_client(self):
        return self.storage_client




import uuid
import random
import unittest
import os
import hashlib
import six
import tempfile
from timeout_decorator import timeout

from .completion_service import CompletionService

from studio.util import has_aws_credentials, filehash
from studio.util import download_file, rand_string

_file_url = 'https://s3-us-west-2.amazonaws.com/ml-enn/' + \
            'deepbilevel_datafiles/' + \
            'mightyai_combined_vocab/mightyai_miscfiles.tar.gz'

_file_s3 = 's3://s3-us-west-2.amazonaws.com/studioml-test/t.txt'

LOCAL_TEST_TIMEOUT = 200
CLOUD_TEST_TIMEOUT = 800


class CompletionServiceTest(unittest.TestCase):

    def _run_test(self,
                  args=None,
                  files={},
                  jobfile=None,
                  expected_results=None,
                  **csargs):

        if not(any(csargs)):
            return

        jobfile = self.get_jobfile(jobfile or 'completion_service_testfunc.py')

        args = args or [0, 1]

        expected_results = expected_results or args
        submission_indices = {}
        n_experiments = len(args)
        experimentId = str(uuid.uuid4())

        with CompletionService(experimentId, **csargs) as cs:
            for i in range(0, n_experiments):
                key = cs.submitTaskWithFiles(jobfile, args[i], files)
                submission_indices[key] = i

            for i in range(0, n_experiments):
                result = cs.getResults(blocking=True)
                self.assertEquals(
                    result[1],
                    expected_results[submission_indices[result[0]]]
                )

    def _run_test_files(self,
                        files,
                        n_experiments=2,
                        **csargs):

        expected_results = [(i, self._get_file_hashes(files))
                            for i in range(n_experiments)]
        args = range(n_experiments)
        self._run_test(
            args=args,
            files=files,
            jobfile='completion_service_testfunc_files.py',
            expected_results=expected_results,
            **csargs
        )

    def _run_test_myfiles(self,
                          n_experiments=2,
                          **csargs):

        mypath = os.path.dirname(os.path.realpath(__file__))

        files_in_workspace = os.listdir(mypath)
        files = {f: os.path.join(mypath, f) for f in files_in_workspace if
                 os.path.isfile(os.path.join(mypath, f))}

        files['url'] = _file_url

        # TODO peterz enable passing aws credentials to google workers
        # if has_aws_credentials():
        #     files['s3'] = _file_s3

        expected_results = [
            (i, self._get_file_hashes(files)) for i in range(n_experiments)
        ]

        args = range(n_experiments)
        self._run_test(
            args=args,
            files=files,
            jobfile='completion_service_testfunc_files.py',
            expected_results=expected_results,
            **csargs
        )

    def _get_file_hashes(self, files):
        retval = {}
        for k, v in six.iteritems(files):
            if '://' in v:
                tmpfilename = os.path.join(
                    tempfile.gettempdir(), rand_string(10))
                download_file(v, tmpfilename)
                retval[k] = filehash(tmpfilename, hashobj=hashlib.md5())
                os.remove(tmpfilename)
            else:
                retval[k] = filehash(v, hashobj=hashlib.md5())

        return retval

    @unittest.skipIf(not has_aws_credentials(),
                     'AWS credentials needed for this test')
    @timeout(CLOUD_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_ec2(self):
        self._run_test(
            config=self.get_config_path(),
            cloud_timeout=100,
            cloud='ec2')

    @unittest.skipIf(not has_aws_credentials(),
                     'AWS credentials needed for this test')
    @timeout(CLOUD_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_ec2spot(self):
        self._run_test_myfiles(
            n_experiments=2,
            config=self.get_config_path(),
            cloud_timeout=100,
            cloud='ec2spot',
        )

    @timeout(LOCAL_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_apiserver(self):
        self._run_test_myfiles(
            n_experiments=2,
            config=self.get_config_path(),
            cloud_timeout=LOCAL_TEST_TIMEOUT
        )

    @unittest.skipIf(True,
                     'Test requires k8s style configurations ' +
                     'that are not yet in play')
    @timeout(LOCAL_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_datacenter(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'datacenter_config.yaml')

        files_in_workspace = os.listdir(mypath)
        files = {f: os.path.join(mypath, f) for f in files_in_workspace if
                 os.path.isfile(os.path.join(mypath, f))}

        files['url'] = _file_url
        files['s3'] = _file_s3

        with get_local_queue_lock():
            self._run_test_files(
                n_experiments=2,
                files=files,
                config=config_path)

    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
        'Need GOOGLE_APPLICATION_CREDENTIALS env variable to' +
        'use google cloud')
    @timeout(CLOUD_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_gcspot(self):
        self._run_test_myfiles(
            n_experiments=2,
            config=self.get_config_path(),
            cloud='gcspot')

    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS_DC' not in os.environ.keys(),
        'Need GOOGLE_APPLICATION_CREDENTIALS_DC env variable to' +
        'use google cloud')
    @timeout(CLOUD_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_datacenter(self):
        oldcred = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = \
            os.environ['GOOGLE_APPLICATION_CREDENTIALS_DC']

        queue_name = 'test_queue_' + str(uuid.uuid4())

        self._run_test_myfiles(
            config=self.get_config_path('test_config_datacenter.yaml'),
            queue=queue_name,
            shutdown_del_queue=True
        )
        if oldcred:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = oldcred

    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
        'Need GOOGLE_APPLICATION_CREDENTIALS env variable to' +
        'use google cloud')
    @timeout(CLOUD_TEST_TIMEOUT, use_signals=False)
    def test_two_experiments_gcloud(self):
        self._run_test_myfiles(
            n_experiments=2,
            config=self.get_config_path(),
            cloud='gcloud')

    @timeout(LOCAL_TEST_TIMEOUT, use_signals=False)
    def test_studiolink(self):

        experiment_id = str(uuid.uuid4())
        arg1 = random.randint(0, 10000)

        jobfile = self.get_jobfile('completion_service_testfunc_saveload.py')
        with CompletionService(experiment_id,
                               config=self.get_config_path(),
                               cloud_timeout=LOCAL_TEST_TIMEOUT) as cs:
            key1 = cs.submitTask(jobfile, arg1)
            ret_key1, result1 = cs.getResults()
            self.assertEquals(key1, ret_key1)
            self.assertEquals(result1, arg1)

            files = {
                'model': 'studio://{}/modeldir'.format(key1)
            }

            key2 = cs.submitTaskWithFiles(jobfile, None, files=files)
            ret_key2, result2 = cs.getResults()
            self.assertEquals(key2, ret_key2)
            self.assertEquals(result2, arg1 + 1)

    @timeout(LOCAL_TEST_TIMEOUT, use_signals=False)
    def test_restart(self):
        experiment_id = str(uuid.uuid4())
        arg1 = random.randint(0, 10000)

        jobfile = self.get_jobfile('completion_service_testfunc_saveload.py')
        with CompletionService(experiment_id,
                               config=self.get_config_path(),
                               cloud_timeout=LOCAL_TEST_TIMEOUT) as cs:
            key1 = cs.submitTask(jobfile, arg1, job_id=0)
            ret_key1, result1 = cs.getResults()
            self.assertEquals(key1, ret_key1)
            self.assertEquals(result1, arg1)

            key2 = cs.submitTask(jobfile, None, job_id=0)
            ret_key2, result2 = cs.getResults()
            self.assertEquals(key2, ret_key2)
            self.assertEquals(result2, arg1 + 1)

    def get_config_path(self, config_name='test_config_http_client.yaml'):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config_http_client.yaml')

        return config_path

    def get_jobfile(self, filename='completion_service_testfunc.py'):
        mypath = os.path.dirname(os.path.realpath(__file__))
        jobfile = os.path.join(
            mypath, filename)

        return jobfile


if __name__ == '__main__':
    unittest.main()



import importlib
import shutil
import pickle
import os
import sys
import six

from studio import fs_tracker, model, logs, util

logger = logs.getLogger('completion_service_client')
try:
    logger.setLevel(model.parse_verbosity(sys.argv[1]))
except BaseException:
    logger.setLevel(10)


def main():
    logger.debug('copying and importing client module')
    logger.debug('getting file mappings')

    artifacts = fs_tracker.get_artifacts()
    files = {}
    logger.debug("Artifacts = {}".format(artifacts))

    for tag, path in six.iteritems(artifacts):
        if tag not in {'workspace', 'modeldir', 'tb', '_runner'}:
            if os.path.isfile(path):
                files[tag] = path
            elif os.path.isdir(path):
                dirlist = os.listdir(path)
                if any(dirlist):
                    files[tag] = os.path.join(
                        path,
                        dirlist[0]
                    )

    logger.debug("Files = {}".format(files))
    script_path = files['clientscript']
    retval_path = fs_tracker.get_artifact('retval')
    util.rm_rf(retval_path)

    # script_name = os.path.basename(script_path)
    new_script_path = os.path.join(os.getcwd(), '_clientscript.py')
    shutil.copy(script_path, new_script_path)

    script_path = new_script_path
    logger.debug("script path: " + script_path)

    mypath = os.path.dirname(script_path)
    sys.path.append(mypath)
    # os.path.splitext(os.path.basename(script_path))[0]
    module_name = '_clientscript'

    client_module = importlib.import_module(module_name)
    logger.debug('loading args')

    args_path = files['args']

    with open(args_path, 'rb') as f:
        args = pickle.loads(f.read())

    logger.debug('calling client function')
    retval = client_module.clientFunction(args, files)

    logger.debug('saving the return value')
    if os.path.isdir(fs_tracker.get_artifact('clientscript')):
        # on go runner:
        logger.debug("Running in a go runner, creating {} for retval"
                     .format(retval_path))
        try:
            os.mkdir(retval_path)
        except OSError:
            logger.debug('retval dir present')

        retval_path = os.path.join(retval_path, 'retval')
        logger.debug("New retval_path is {}".format(retval_path))

    logger.debug('Saving retval')
    with open(retval_path, 'wb') as f:
        f.write(pickle.dumps(retval, protocol=2))
    logger.debug('Done')


if __name__ == "__main__":
    main()



import os
import sys
import subprocess
import argparse
import json
import psutil
import time
import six
from pygtail import Pygtail
import threading

from apscheduler.schedulers.background import BackgroundScheduler

from . import fs_tracker, model, logs
from .local_queue import LocalQueue
from .gpu_util import get_available_gpus, get_gpu_mapping, get_gpus_summary
from .experiment import Experiment
from .util import sixdecode, str2duration, retry
from .model import parse_verbosity

logs.getLogger('apscheduler.scheduler').setLevel(logs.ERROR)


class LocalExecutor(object):
    """Runs job while capturing environment and logs results.
    """

    def __init__(self, args):
        self.config = args.config

        if args.guest:
            self.config['database']['guest'] = True

        self.logger = logs.getLogger('LocalExecutor')
        self.logger.setLevel(model.parse_verbosity(self.config.get('verbose')))
        self.logger.debug("Config: ")
        self.logger.debug(self.config)

    def run(self, experiment):
        if isinstance(experiment, six.string_types):
            experiment = self.db.get_experiment(experiment)
        elif not isinstance(experiment, Experiment):
            raise ValueError("Unknown type of experiment: " +
                             str(type(experiment)))

        self.logger.info("Experiment key: " + experiment.key)

        with model.get_db_provider(self.config) as db:
            db.start_experiment(experiment)

            """ Override env variables with those inside the queued message
            """
            env = dict(os.environ)
            if 'env' in self.config.keys():
                for k, v in six.iteritems(self.config['env']):
                    if v is not None:
                        env[str(k)] = str(v)

            env['PYTHONUNBUFFERED'] = 'TRUE'

            fs_tracker.setup_experiment(env, experiment, clean=False)
            log_path = fs_tracker.get_artifact_cache('output', experiment.key)

            # log_path = os.path.join(model_dir, self.config['log']['name'])

            self.logger.debug('Child process environment:')
            self.logger.debug(str(env))

            sched = BackgroundScheduler()
            sched.start()

            with open(log_path, 'w') as output_file:
                python = 'python'
                if experiment.pythonver[0] == '3':
                    python = 'python3'

                python = which(python)

                cmd = [python, experiment.filename] + experiment.args
                cwd = experiment.artifacts['workspace']['local']
                container_artifact = experiment.artifacts.get('_singularity')
                if container_artifact:
                    container = container_artifact.get('local')
                    if not container:
                        container = container_artifact.get('qualified')

                    cwd = fs_tracker.get_artifact_cache(
                        'workspace', experiment.key)

                    for tag, art in six.iteritems(experiment.artifacts):
                        local_path = art.get('local')
                        if not art['mutable'] and os.path.exists(local_path):
                            os.symlink(
                                art['local'],
                                os.path.join(os.path.dirname(cwd), tag)
                            )

                    if experiment.filename is not None:
                        cmd = [
                            'singularity',
                            'exec',
                            container,
                        ] + cmd
                    else:
                        cmd = ['singularity', 'run', container]

                self.logger.info('Running cmd: {} in {}'.format(cmd, cwd))

                p = subprocess.Popen(
                    cmd,
                    stdout=output_file,
                    stderr=subprocess.STDOUT,
                    env=env,
                    cwd=cwd
                )

                # simple hack to show what's in the log file
                # ptail = subprocess.Popen(["tail", "-f", log_path])

                logtail = Pygtail(log_path)

                def tail_func():
                    while logtail:
                        for line in logtail:
                            print(line)

                        time.sleep(0.1)

                tail_thread = threading.Thread(target=tail_func)
                tail_thread.start()

                minutes = 0
                if self.config.get('saveWorkspaceFrequency'):
                    minutes = int(
                        str2duration(
                            self.config['saveWorkspaceFrequency'])
                        .total_seconds() / 60)

                def checkpoint():
                    try:
                        db.checkpoint_experiment(experiment)
                    except BaseException as e:
                        self.logger.info(e)

                sched.add_job(
                    checkpoint,
                    'interval',
                    minutes=minutes)

                metrics_path = fs_tracker.get_artifact_cache(
                    '_metrics', experiment.key)

                minutes = 0
                if self.config.get('saveMetricsFrequency'):
                    minutes = int(
                        str2duration(
                            self.config['saveMetricsFrequency'])
                        .total_seconds() / 60)

                sched.add_job(
                    lambda: save_metrics(metrics_path),
                    'interval',
                    minutes=minutes)

                def kill_if_stopped():
                    db_expr = db.get_experiment(
                        experiment.key,
                        getinfo=False)

                    # Transient issues with getting experiment data might
                    # result in a None value being returned, as result
                    # leave the experiment running because we wont be able to
                    # do anything else even if this experiment is stopped
                    # in any event if the experiment runs too long then it
                    # will exceed its allocated time and stop
                    if db_expr is not None:
                        if db_expr.status == 'stopped':
                            p.kill()
                            return

                    if experiment.max_duration is not None and \
                            time.time() > experiment.time_started + \
                            int(str2duration(experiment.max_duration)
                                .total_seconds()):

                        p.kill()

                sched.add_job(kill_if_stopped, 'interval', seconds=10)

                try:
                    p.wait()
                finally:
                    save_metrics(metrics_path)
                    sched.shutdown()
                    logtail = None
                    db.checkpoint_experiment(experiment)
                    db.finish_experiment(experiment)
                    return p.returncode


def allocate_resources(experiment, config=None, verbose=10):
    logger = logs.getLogger('allocate_resources')
    logger.setLevel(verbose)
    logger.info('Allocating resources {} for experiment {}'
                .format(experiment.resources_needed, experiment.key))

    ret_val = True
    gpus_needed = int(experiment.resources_needed.get('gpus')) \
        if experiment.resources_needed else 0

    if gpus_needed > 0:
        ret_val = ret_val and allocate_gpus(gpus_needed,
                                            experiment.resources_needed,
                                            config)
    else:
        allocate_gpus(0)

    return ret_val


def allocate_gpus(gpus_needed, resources_needed={}, config=None):
    # Only disable gpus if gpus_needed < 0
    if gpus_needed < 0:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
        return True
    elif gpus_needed == 0:
        return True

    gpu_mem_needed = resources_needed.get('gpuMem', None)
    strict = resources_needed.get('gpuMemStrict', False)

    available_gpus = get_available_gpus(gpu_mem_needed, strict)
    gpu_mapping = get_gpu_mapping()
    mapped_gpus = [str(gpu_mapping[g]) for g in available_gpus]

    os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
    if len(mapped_gpus) >= gpus_needed:
        os.environ['CUDA_VISIBLE_DEVICES'] = ','.join(
            mapped_gpus[:gpus_needed])
        return True
    else:
        return False


def main(args=sys.argv):
    parser = argparse.ArgumentParser(
        description='Studio worker. \
                     Usage: studio-local-worker \
                     ')

    parser.add_argument('--config', help='configuration file', default=None)
    parser.add_argument(
        '--guest',
        help='Guest mode (does not require db credentials)',
        action='store_true')
    parser.add_argument(
        '--timeout',
        default=0, type=int)
    parser.add_argument(
        '--verbose',
        default='error')

    parsed_args, script_args = parser.parse_known_args(args)
    verbose = parse_verbosity(parsed_args.verbose)

    queue = LocalQueue(verbose=verbose)
    # queue = glob.glob(fs_tracker.get_queue_directory() + "/*")
    # wait_for_messages(queue, parsed_args.timeout)
    returncode = worker_loop(queue, parsed_args, timeout=parsed_args.timeout)
    sys.exit(returncode)


def which(program):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def worker_loop(queue, parsed_args,
                single_experiment=False,
                timeout=0,
                verbose=None):

    fetch_artifacts = True

    logger = logs.getLogger('worker_loop')

    hold_period = 4
    retval = 0
    while True:
        msg = queue.dequeue(acknowledge=False, timeout=timeout)
        if not msg:
            break

        # first_exp, ack_key = queue.dequeue(acknowledge=False)
        first_exp, ack_key = msg

        data_dict = json.loads(sixdecode(first_exp))
        experiment_key = data_dict['experiment']['key']
        config = data_dict['config']

        parsed_args.config = config
        if verbose:
            config['verbose'] = verbose
        else:
            verbose = model.parse_verbosity(config.get('verbose'))

        logger.setLevel(verbose)

        logger.debug('Received message: \n{}'.format(data_dict))

        executor = LocalExecutor(parsed_args)

        with model.get_db_provider(config) as db:
            # experiment = experiment_from_dict(data_dict['experiment'])
            def try_get_experiment():
                experiment = db.get_experiment(experiment_key)
                if experiment is None:
                    raise ValueError(
                        'experiment is not found - indicates storage failure')
                return experiment

            experiment = retry(
                try_get_experiment,
                sleep_time=10,
                logger=logger)

            if config.get('experimentLifetime') and \
                int(str2duration(config['experimentLifetime'])
                    .total_seconds()) + experiment.time_added < time.time():
                logger.info(
                    'Experiment expired (max lifetime of {} was exceeded)'
                    .format(config.get('experimentLifetime'))
                )
                queue.acknowledge(ack_key)
                continue

            if allocate_resources(experiment, config, verbose=verbose):
                def hold_job():
                    queue.hold(ack_key, hold_period)

                hold_job()
                sched = BackgroundScheduler()
                sched.add_job(hold_job, 'interval', minutes=hold_period / 2)
                sched.start()

                try:
                    python = 'python'
                    if experiment.pythonver[0] == '3':
                        python = 'python3'
                    if '_singularity' not in experiment.artifacts.keys():
                        pip_diff = pip_needed_packages(
                            experiment.pythonenv, python)
                        if any(pip_diff):
                            logger.info(
                                'Setting up python packages for experiment')
                            if pip_install_packages(
                                    pip_diff,
                                    python,
                                    logger
                            ) != 0:

                                logger.info(
                                    "Installation of all packages together " +
                                    " failed, "
                                    "trying one package at a time")

                                for pkg in pip_diff:
                                    pip_install_packages([pkg], python, logger)

                    for tag, art in six.iteritems(experiment.artifacts):
                        if fetch_artifacts or 'local' not in art.keys():
                            logger.info('Fetching artifact ' + tag)
                            if tag == 'workspace':
                                art['local'] = retry(lambda: db.get_artifact(
                                    art,
                                    only_newer=False),
                                    sleep_time=10,
                                    logger=logger)
                            else:
                                art['local'] = retry(
                                    lambda: db.get_artifact(art),
                                    sleep_time=10,
                                    logger=logger
                                )

                    returncode = executor.run(experiment)
                    if returncode != 0:
                        retval = returncode
                finally:
                    sched.shutdown()
                    queue.acknowledge(ack_key)

                if single_experiment:
                    logger.info('single_experiment is True, quitting')
                    return retval
            else:
                logger.info('Cannot run experiment ' + experiment.key +
                            ' due lack of resources. Will retry')
                # Debounce failed requests we cannot service yet
                time.sleep(config.get('sleep_time', 5))

        # wait_for_messages(queue, timeout, logger)

        # queue = glob.glob(fs_tracker.get_queue_directory() + "/*")

    logger.info("Queue in {} is empty, quitting"
                .format(fs_tracker.get_queue_directory()))

    return retval


def pip_install_packages(packages, python='python', logger=None):
    pipp = subprocess.Popen(
        [python, '-m', 'pip', 'install'] + [p for p in packages],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    pipout, _ = pipp.communicate()
    pipout = pipout.decode('utf-8')
    # return pip.main(['install'] + list(packages))

    if logger:
        logger.info("pip output: \n" + pipout)
    return pipp.returncode


def wait_for_messages(queue, timeout, logger=None):
    wait_time = 0
    wait_step = 5
    timeout = int(timeout)
    if timeout == 0:
        return

    while not queue.has_next():
        if logger:
            logger.info(
                'No messages found, sleeping for {} s (total wait time {} s)'
                .format(wait_step, wait_time))
        time.sleep(wait_step)
        wait_time += wait_step
        if timeout > 0 and timeout < wait_time:
            if logger:
                logger.info('No jobs found in the queue during {} s'.
                            format(timeout))
            return


def save_metrics(path):
    cpu_load = psutil.cpu_percent()
    cpu_mem = psutil.virtual_memory().used
    timestamp = time.time()
    with open(path, 'a') as f:
        entry = 'time: {} CPU: {} mem: {} {} \n' \
                .format(
                    timestamp,
                    cpu_load,
                    cpu_mem,
                    get_gpus_summary())

        f.write(entry)


def pip_needed_packages(packages, python='python'):

    pipp = subprocess.Popen(
        [python, '-m', 'pip', 'freeze'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)

    pipout, _ = pipp.communicate()
    pipout = pipout.decode('utf-8')
    current_packages = {l.strip() for l in pipout.strip().split('\n')}

    # current_packages = {p._key + '==' + p._version for p in
    #                    pip.pip.get_installed_distributions(local_only=True)}

    return {p for p in packages} - current_packages


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import pika
import json
import time
import logging
import threading

from . import logs


class RMQueue(object):
    """This publisher will handle failures and closures and will
    attempt to restart things

    delivery confirmations are used to track messages that have
    been sent and if they have been confirmed

    """

    def __init__(self, queue, route, amqp_url='',
                 config=None, logger=None, verbose=None):
        """Setup the example publisher object, passing in the URL we will use
        to connect to RabbitMQ.
        """
        self._rmq_lock = threading.RLock()
        self._connection = None
        self._channel = None
        self._consumer = None
        self._consume_ready = False

        self._msg_tracking_lock = threading.RLock()
        self._deliveries = []
        self._acked = 0
        self._nacked = 0
        self._message_number = 0

        self._rmq_msg = None
        self._rmq_id = None

        self._stopping = False
        self._exchange = 'StudioML.topic'
        self._exchange_type = 'topic'
        self._routing_key = route

        self._url = amqp_url

        if logger is not None:
            self._logger = logger
        else:
            self._logger = logs.getLogger('RabbitMQ')
            self._logger.setLevel(logging.INFO)

        if config is not None:
            # extract from the config data structure any settings related to
            # queue messaging for rabbit MQ
            if 'cloud' in config:
                if 'queue' in config['cloud']:
                    if 'rmq' in config['cloud']['queue']:
                        self._url = config['cloud']['queue']['rmq']
                        self._logger.warn('url {}'
                                          .format(self._url))

        self._queue = queue

        # The pika library for RabbitMQ has an asynchronous run method
        # that needs to run forever and will do reconnections etc
        # automatically for us
        thr = threading.Thread(target=self.run, args=(), kwargs={})
        thr.setDaemon(True)
        thr.start()

    def connect(self):
        """
        When the connection is established, the on_connection_open method
        will be invoked by pika. If you want the reconnection to work, make
        sure you set stop_ioloop_on_close to False, which is not the default
        behavior of this adapter.

        :rtype: pika.SelectConnection

        """
        params = pika.URLParameters(self._url)
        return pika.SelectConnection(
            params,
            on_open_callback=self.on_connection_open,
            on_close_callback=self.on_connection_closed,
            stop_ioloop_on_close=False)

    def on_connection_open(self, unused_connection):
        """
        :type unused_connection: pika.SelectConnection
        """
        self.open_channel()

    def on_connection_closed(self, connection, reply_code, reply_text):
        """
        on any close reconnect to RabbitMQ, until the stopping is set

        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given

        """
        with self._rmq_lock:
            self._channel = None
            if self._stopping:
                self._connection.ioloop.stop()
            else:
                # retry in 5 seconds
                self._logger.info('connection closed, retry in 5 seconds: ' +
                                  str(reply_code) + ' ' + reply_text)
                self._connection.add_timeout(5, self._connection.ioloop.stop)

    def open_channel(self):
        """
        open a new channel using the Channel.Open RPC command. RMQ confirms
        the channel is open by sending the Channel.OpenOK RPC reply, the
        on_channel_open method will be invoked.
        """
        self._logger.debug('creating a new channel')

        with self._rmq_lock:
            self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """
        on channel open, declare the exchange to use

        :param pika.channel.Channel channel: The channel object

        """
        self._logger.debug('created a new channel')

        with self._rmq_lock:
            self._channel = channel
            self._channel.basic_qos(prefetch_count=0)
            self._channel.add_on_close_callback(self.on_channel_closed)

        self.setup_exchange(self._exchange)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """
        physical network issues and logical protocol abuses can
        result in a closure of the channel.

        :param pika.channel.Channel channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed

        """
        self._logger.info(
            'channel closed ' +
            str(reply_code) +
            ' ' +
            reply_text)
        with self._rmq_lock:
            self._channel = None
            if not self._stopping:
                self._connection.close()

    def setup_exchange(self, exchange_name):
        """
        exchange setup by invoking the Exchange.Declare RPC command.
        When complete, the on_exchange_declareok method will be invoked
        by pika.

        :param str|unicode exchange_name: The name of the exchange to declare

        """
        self._logger.debug('declaring exchange ' + exchange_name)
        with self._rmq_lock:
            self._channel.exchange_declare(callback=self.on_exchange_declareok,
                                           exchange=exchange_name,
                                           exchange_type=self._exchange_type,
                                           durable=True,
                                           auto_delete=True)

    def on_exchange_declareok(self, unused_frame):
        """
        completion callback for the Exchange.Declare RPC command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response

        """
        self._logger.debug('declared exchange ' + self._exchange)
        self.setup_queue(self._queue)

    def setup_queue(self, queue_name):
        """
        Setup the queue invoking the Queue.Declare RPC command.
        The completion callback is, the on_queue_declareok method.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        self._logger.debug('declare queue ' + queue_name)
        with self._rmq_lock:
            self._channel.queue_declare(self.on_queue_declareok, queue_name)

    def on_queue_declareok(self, method_frame):
        """
        Queue.Declare RPC completion callback.
        In this method the queue and exchange are bound together
        with the routing key by issuing the Queue.Bind
        RPC command.

        The completion callback is the on_bindok method.

        :param pika.frame.Method method_frame: The Queue.DeclareOk frame

        """
        self._logger.debug(
            'binding ' +
            self._exchange +
            ' to ' +
            self._queue +
            ' with ' +
            self._routing_key)
        with self._rmq_lock:
            self._channel.queue_bind(self.on_bindok, self._queue,
                                     self._exchange, self._routing_key)

    def on_bindok(self, unused_frame):
        """This method is invoked by pika when it receives the Queue.BindOk
        response from RabbitMQ. Since we know we're now setup and bound, it's
        time to start publishing."""
        self._logger.info(
            'bound ' +
            self._exchange +
            ' to ' +
            self._queue +
            ' with ' +
            self._routing_key)

        """
        Send the Confirm.Select RPC method to RMQ to enable delivery
        confirmations on the channel. The only way to turn this off is to close
        the channel and create a new one.

        When the message is confirmed from RMQ, the
        on_delivery_confirmation method will be invoked passing in a Basic.Ack
        or Basic.Nack method from RMQ that will indicate which messages it
        is confirming or rejecting.
        """
        with self._rmq_lock:
            self._channel.confirm_delivery(self.on_delivery_confirmation)

    def on_delivery_confirmation(self, method_frame):
        """
        RMQ callback for responses to a Basic.Publish RPC
        command, passing in either a Basic.Ack or Basic.Nack frame with
        the delivery tag of the message that was published. The delivery tag
        is an integer counter indicating the message number that was sent
        on the channel via Basic.Publish. Here we're just doing house keeping
        to keep track of stats and remove message numbers that we expect
        a delivery confirmation of from the list used to keep track of messages
        that are pending confirmation.

        :param pika.frame.Method method_frame: Basic.Ack or Basic.Nack frame

        """
        confirmation_type = method_frame.method.NAME.split('.')[1].lower()
        self._logger.debug('received ' +
                           confirmation_type +
                           ' for delivery tag: ' +
                           str(method_frame.method.delivery_tag))

        with self._msg_tracking_lock:
            if confirmation_type == 'ack':
                self._acked += 1
            elif confirmation_type == 'nack':
                self._nacked += 1
            self._deliveries.remove(method_frame.method.delivery_tag)
            self._logger.info('published ' +
                              str(self._message_number) +
                              ' messages, ' +
                              str(len(self._deliveries)) +
                              ' have yet to be confirmed, ' +
                              str(self._acked) +
                              ' were acked and ' +
                              str(self._nacked) +
                              ' were nacked')

    def run(self):
        """
        Blocking run loop, connecting and then starting the IOLoop.
        """
        self._logger.info('RMQ started')
        while not self._stopping:
            self._connection = None
            with self._msg_tracking_lock:
                self._deliveries = []
                self._acked = 0
                self._nacked = 0
                self._message_number = 0

            try:
                with self._rmq_lock:
                    self._connection = self.connect()
                self._logger.info('RMQ connected')
                self._connection.ioloop.start()
            except KeyboardInterrupt:
                self.stop()
                if (self._connection is not None and
                        not self._connection.is_closed):
                    # Finish closing
                    self._connection.ioloop.start()

        self._logger.info('RMQ stopped')

    def stop(self):
        """
        Stop the by closing the channel and connection and setting
        a stop state.

        The IOLoop is started independently which means we need this
        method to handle things such as the Try/Catch when KeyboardInterrupts
        are caught.
        Starting the IOLoop again will allow the publisher to cleanly
        disconnect from RMQ.
        """
        self._logger.info('stopping')
        self._stopping = True
        self.close_channel()
        self.close_connection()

    def close_channel(self):
        """
        Close channel by sending the Channel.Close RPC command.
        """
        with self._rmq_lock:
            if self._channel is not None:
                self._logger.info('closing the channel')
                self._channel.close()

    def close_connection(self):
        with self._rmq_lock:
            if self._connection is not None:
                self._logger.info('closing connection')
                self._connection.close()

    def clean(self, timeout=0):
        while True:
            msg = self.dequeue(timeout=timeout)
            if not msg:
                break
        return

    def get_name(self):
        return self._queue

    def enqueue(self, msg, retries=10):
        """
        Publish a message to RMQ, appending a list of deliveries with
        the message number that was sent.  This list will be used to
        check for delivery confirmations in the
        on_delivery_confirmations method.
        """
        if self._url is None:
            raise Exception('url for rmq not initialized')

        if msg is None:
            raise Exception(
                'message was None, it needs a meaningful value to be sent')

        # Wait to see if the channel gets opened
        tries = retries
        while tries != 0:
            if self._channel is None:
                self._logger.warn(
                    'failed to send message ({} tries left) to {} as '
                    'the channel API was not initialized' .format(
                        tries, self._url))
            elif not self._channel.is_open:
                self._logger.warn(
                    'failed to send message ({} tries left) to {} as '
                    'the channel was not open' .format(
                        tries, self._url))
            else:
                break

            time.sleep(1)
            tries -= 1

        if tries == 0:
            raise Exception('studioml request could not be sent')

        self._logger.debug('sending message {} to {} '
                           .format(msg, self._url))
        properties = pika.BasicProperties(app_id='studioml',
                                          content_type='application/json')

        self._channel.basic_publish(exchange=self._exchange,
                                    routing_key=self._routing_key,
                                    body=msg,
                                    properties=properties,
                                    mandatory=True)
        self._logger.debug('sent message to {} '
                           .format(self._url))

        message_number = 0
        with self._msg_tracking_lock:
            self._message_number += 1

            message_number = self._message_number
            self._deliveries.append(self._message_number)

        tries = retries
        while tries != 0:
            time.sleep(1)

            with self._msg_tracking_lock:
                if message_number not in self._deliveries:
                    self._logger.debug('sent message acknowledged to {} ' +
                                       'after waiting {} seconds'
                                       .format(self._url, abs(tries - 5)))

                    return message_number
                else:
                    tries -= 1

        raise Exception('studioml message was never acknowledged to {} ' +
                        'after waiting {} seconds'
                        .format(self._url, abs(tries - 5)))

    def dequeue(self, acknowledge=True, timeout=0):
        msg = None

        # start the consumer and allow single messages to returned via
        # this method to the caller blocking using a callback lock
        # while waiting
        for i in range(timeout + 1):
            with self._rmq_lock:
                if self._consumer is None and self._channel is not None:
                    self._consumer = \
                        self._channel.basic_consume(self.on_message,
                                                    queue=self._queue)

                if self._rmq_msg is not None:
                    self._logger.info('message {} from {} '
                                      .format(self._rmq_msg, self._url))
                    return self._rmq_msg, self._rmq_id
                else:
                    self._logger.info('idle {} {}'
                                      .format(self._url, self._queue))

            if i >= timeout:
                self._logger.info('timed-out')
                return None

            time.sleep(1)

        self._logger.info('dequeue done')

    def on_message(self, unused_channel, basic_deliver, properties, body):

        with self._rmq_lock:
            if self._channel is not None:
                # Cancel the consumer as we only consume 1 message
                # at a time
                self._channel.basic_cancel(
                    nowait=True, consumer_tag=self._consumer)
            self._consumer = None

            # If we already had a delivered message, reject the one we just got
            if self._rmq_msg is not None:
                if self._connection is not None:
                    self._channel.basic_nack(
                        delivery_tag=basic_deliver.delivery_tag)
            else:
                self._rmq_msg = body
                self._rmq_id = basic_deliver.delivery_tag

    def has_next(self):
        raise NotImplementedError(
            'using has_next with distributed queue is not supportable')

    def acknowledge(self, ack_id):
        with self._rmq_lock:
            self._rmq_msg = None
            self._rmq_id = None
            if self._channel is None:
                return None
            result = self._channel.basic_ack(delivery_tag=ack_id)

    def hold(self, ack_id, minutes):
        # Nothing is needed here as the message will remain while the channel
        # remains open, or we nack it
        pass

    def delete(self):
        raise NotImplementedError('')




DB_KEY = "database"
STORE_KEY = "store"

# Global dictionary which keeps Database Provider
# and Artifact Store objects created from experiment configuration.
_model_setup = None

def setup_model(db_provider, artifact_store):
    _model_setup = { DB_KEY: db_provider, STORE_KEY: artifact_store }

def get_db_provider():
    if _model_setup is None:
        return None
    return _model_setup.get(DB_KEY, None)

def get_artifact_store():
    if _model_setup is None:
        return None
    return _model_setup.get(STORE_KEY, None)
